# ===================================================
# tg01ad tests

import unittest
from slycot import transform
import numpy as np

from numpy.testing import assert_raises, assert_almost_equal, assert_equal

# test1 input parameters

test1_l = 4
test1_n = 4
test1_m = 2
test1_p = 2
test1_job = 'A'
test1_thresh = 0.0

test1_A = \
    np.array([[-1.0,    0.0,    0.0,    3e-3   ],
              [ 0.0,    0.0,    0.1,    2e-2   ],
              [ 1e2,    10.0,   0.0,    0.4    ],
              [ 0.0,    0.0,    0.0,    0.0    ]])

test1_E = \
    np.array([[ 1.0,    0.2,    0.0,    0.0    ],
              [ 0.0,    1.0,    0.0,    1e-2   ],
              [ 3e2,    90.0,   6.0,    0.3    ],
              [ 0.0,    0.0,    20.0,   0.0    ]])

test1_B = \
    np.array([[ 10.0,   0.0    ],
              [ 0.0,    0.0    ],
              [ 0.0,    1e3    ],
              [ 1e4,    1e4    ]])

test1_C = \
    np.array([[-0.1,    0.0,    1e-3,    0.0   ],
              [ 0.0,    1e-2,  -1e-3,    1e-4  ]])

test1_A_desired = \
    np.array([[-1.0,    0.0,    0.0,     0.3   ],
              [ 0.0,    0.0,    1.0,     2.0   ],
              [ 1.0,    0.1,    0.0,     0.4   ],
              [ 0.0,    0.0,    0.0,     0.0   ]])

test1_E_desired = \
    np.array([[ 1.0,    0.2,    0.0,     0.0   ],
              [ 0.0,    1.0,    0.0,     1.0   ],
              [ 3.0,    0.9,    0.6,     0.3   ],
              [ 0.0,    0.0,    0.2,     0.0   ]])

test1_B_desired = \
    np.array([[ 1e2,    0.0    ],
              [ 0.0,    0.0    ],
              [ 0.0,    1e2    ],
              [ 1e2,    1e2    ]])

test1_C_desired = \
    np.array([[-1e-2,   0.0,    1e-3,    0.0   ],
              [ 0.0,    1e-3,  -1e-3,    1e-3  ]])

test1_lscale_desired = \
    np.array([  10.0,   10.0,   0.1,     1e-2  ])

test1_rscale_desired = \
    np.array([  0.1,    0.1,    1.0,     10.0  ])

class test_tg01ad(unittest.TestCase):
    """ test1: Verify tg01ad with input parameters according to example in documentation """

    def test1_tg01ad(self):

        A,E,B,C,lscale,rscale = transform.tg01ad(l=test1_l,n=test1_n,m=test1_m,p=test1_p,A=test1_A,E=test1_E,B=test1_B,C=test1_C,job=test1_job, thresh=test1_thresh)

        assert_almost_equal(A, test1_A_desired)
        assert_almost_equal(E, test1_E_desired)
        assert_almost_equal(B, test1_B_desired)
        assert_almost_equal(C, test1_C_desired)
        assert_almost_equal(lscale, test1_lscale_desired)
        assert_almost_equal(rscale, test1_rscale_desired)


if __name__ == "__main__":
    unittest.main()



# ===================================================
# tb05ad tests

from slycot import transform
from slycot.exceptions import SlycotArithmeticError, SlycotParameterError
import sys
import numpy as np
from scipy.linalg import matrix_balance, eig
import unittest
from numpy.testing import assert_almost_equal

# set the random seed so we can get consistent results.
np.random.seed(40)
CASES = {}

# This was (pre 2020) a known failure for tb05ad when running job 'AG'
CASES['known'] = {'A': np.array([[-0.5,  0.,  0.,  0.],
                                 [ 0.,  -1.,  0.,  0.],
                                 [ 1.,   0., -0.5, 0.],
                                 [ 0.,   1.,  0., -1.]]),
                  'B': np.array([[ 1.,  0.],
                                 [ 0.,  1.],
                                 [ 0.,  0.],
                                 [ 0.,  0.]]),
                  'C': np.array([[ 0.,  1.,  1.,  0.],
                                 [ 0.,  1.,  0.,  1.],
                                 [ 0.,  1.,  1.,  1.]])}

n = 20
p = 10
m = 14

CASES['pass1'] = {'A': np.random.randn(n, n),
                  'B':  np.random.randn(n, m),
                  'C':  np.random.randn(p, n)}


class test_tb05ad(unittest.TestCase):

    def test_tb05ad_ng(self):
        """
        Test that tb05ad with job 'NG' computes the correct
        frequency response.
        """
        for key in CASES:
            sys = CASES[key]
            self.check_tb05ad_AG_NG(sys, 10*1j, 'NG')

    def test_tb05ad_ag(self):
        """
        Test that tb05ad with job 'AG' computes the correct
        frequency response.
        """
        for key in CASES:
            sys = CASES[key]
            self.check_tb05ad_AG_NG(sys, 10*1j, 'AG')

    def test_tb05ad_nh(self):
        """Test that tb05ad with job = 'NH' computes the correct
        frequency response after conversion to Hessenberg form.

        First call tb05ad with job='NH' to transform to upper Hessenberg
        form which outputs the transformed system.
        Subsequently, call tb05ad with job='NH' using this transformed system.
        """
        jomega = 10*1j
        for key in CASES:
            sys = CASES[key]
            sys_transformed = self.check_tb05ad_AG_NG(sys, jomega, 'NG')
            self.check_tb05ad_NH(sys_transformed, sys, jomega)

    def test_tb05ad_errors(self):
        """
        Test tb05ad error handling. We give wrong inputs and
        and check that this raises an error.
        """
        self.check_tb05ad_errors(CASES['pass1'])

    def check_tb05ad_AG_NG(self, sys, jomega, job):
        """
        Check that tb05ad computes the correct frequency response when
        running jobs 'AG' and/or 'NG'.

        Inputs
        ------

        sys: A a dict of system matrices with keys 'A', 'B', and 'C'.
        jomega: A complex scalar, which is the frequency we are
                evaluating the system at.
        job: A string, either 'AG' or 'NH'

        Returns
        -------
        sys_transformed: A dict of the system matrices which have been
                         transformed according the job.
        """
        n, m = sys['B'].shape
        p = sys['C'].shape[0]
        result = transform.tb05ad(n, m, p, jomega,
                                  sys['A'], sys['B'], sys['C'], job=job)
        g_i = result[3]
        hinvb = np.linalg.solve(np.eye(n) * jomega - sys['A'], sys['B'])
        g_i_solve = sys['C'].dot(hinvb)
        assert_almost_equal(g_i_solve, g_i)
        sys_transformed = {'A': result[0], 'B': result[1], 'C': result[2]}
        return sys_transformed

    def check_tb05ad_NH(self, sys_transformed, sys, jomega):
        """
        Check tb05ad, computes the correct frequency response when
        job='NH' and we supply system matrices 'A', 'B', and 'C'
        which have been transformed by a previous call to tb05ad.
        We check we get the same result as computing C(sI - A)^-1B
        with the original system.

        Inputs
        ------

        sys_transformed: A a dict of the transformed (A in upper
                         hessenberg form) system matrices with keys
                         'A', 'B', and 'C'.

        sys: A dict of the original un-transformed system matrices.

        jomega: A complex scalar, which is the frequency to evaluate at.

        """

        n, m = sys_transformed['B'].shape
        p = sys_transformed['C'].shape[0]
        result = transform.tb05ad(n, m, p, jomega, sys_transformed['A'],
                                  sys_transformed['B'], sys_transformed['C'],
                                  job='NH')
        g_i = result[0]
        hinvb = np.linalg.solve(np.eye(n) * jomega - sys['A'], sys['B'])
        g_i_solve = sys['C'].dot(hinvb)
        assert_almost_equal(g_i_solve, g_i)

    def check_tb05ad_errors(self, sys):
        """
        Check the error handling of tb05ad. We give wrong inputs and
        and check that this raises an error.
        """
        n, m = sys['B'].shape
        p = sys['C'].shape[0]
        jomega = 10*1j
        # test error handling
        # wrong size A
        with self.assertRaises(SlycotParameterError) as cm:
            transform.tb05ad(
                n+1, m, p, jomega, sys['A'], sys['B'], sys['C'], job='NH')
        assert cm.exception.info == -7
        # wrong size B
        with self.assertRaises(SlycotParameterError) as cm:
            transform.tb05ad(
                n, m+1, p, jomega, sys['A'], sys['B'], sys['C'], job='NH')
        assert cm.exception.info == -9
        # wrong size C
        with self.assertRaises(SlycotParameterError) as cm:
            transform.tb05ad(
                n, m, p+1, jomega, sys['A'], sys['B'], sys['C'], job='NH')
        assert cm.exception.info == -11
        # unrecognized job
        with self.assertRaises(SlycotParameterError) as cm:
            transform.tb05ad(
                n, m, p, jomega, sys['A'], sys['B'], sys['C'], job='a')
        assert cm.exception.info == -1

    @unittest.skipIf(sys.version < "3", "no assertRaisesRegex in old Python")
    def test_tb05ad_resonance(self):
        """ Test tb05ad resonance failure.

        Actually test one of the exception messages. These
        are parsed from the docstring, tests both the info index and the
        message
        """
        A = np.array([[0, -1],
                      [1, 0]])
        B = np.array([[1],
                      [0]])
        C = np.array([[0, 1]])
        jomega = 1j
        with self.assertRaisesRegex(
                SlycotArithmeticError,
                r"Either `freq`.* is too near to an eigenvalue of A,\n"
                r"or `rcond` is less than the machine precision EPS.") as cm:
            transform.tb05ad(2, 1, 1, jomega, A, B, C, job='NH')
        assert cm.exception.info == 2

    def test_tb05ad_balance(self):
        """Test balancing in tb05ad.

        Tests for the cause of the problem reported in issue #11
        balancing permutations were not correctly applied to the
        C and D matrix.
        """

        # find a good test case. Some sparsity,
        # some zero eigenvalues, some non-zero eigenvalues,
        # and proof that the 1st step, with dgebal, does some
        # permutation and some scaling
        crit = False
        n = 8
        while not crit:
            A = np.random.randn(n, n)
            A[np.random.uniform(size=(n, n)) > 0.35] = 0.0

            Aeig = eig(A)[0]
            neig0 = np.sum(np.abs(Aeig) == 0)
            As, T = matrix_balance(A)
            nperm = np.sum(np.diag(T == 0))
            nscale = n - np.sum(T == 1.0)
            crit = nperm < n and nperm >= n//2 and \
                neig0 > 1 and neig0 <= 3 and nscale > 0

        # print("number of permutations", nperm, "eigenvalues=0", neig0)
        B = np.random.randn(8, 4)
        C = np.random.randn(3, 8)

        # do a run
        jomega = 1.0
        At, Bt, Ct, rcond, g_jw, ev, hinvb, info = transform.tb05ad(
            8, 4, 3, jomega, A, B, C, job='AG')

        # remove information on Q, in lower sub-triangle part of A
        At = np.triu(At, k=-1)

        # now after the balancing in DGEBAL, and conversion to
        # upper Hessenberg form:
        # At = Q^T * (P^-1 * A * P ) * Q
        # with Q orthogonal
        # Ct = C * P * Q
        # Bt = Q^T * P^-1 * B
        # so test with Ct * At * Bt  ==  C * A * B
        # and verify that eigenvalues of both A matrices are close
        assert_almost_equal(np.dot(np.dot(Ct, At), Bt),
                            np.dot(np.dot(C, A), B))
        # uses a sort, there is no guarantee on the order of eigenvalues
        eigAt = eig(At)[0]
        idxAt = np.argsort(eigAt)
        eigA = eig(A)[0]
        idxA = np.argsort(eigA)
        assert_almost_equal(eigA[idxA], eigAt[idxAt])


if __name__ == "__main__":
    unittest.main()



#
#       math.py
#
#       Copyright 2010 Enrico Avventi <avventi@Lonewolf>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License version 2 as
#       published by the Free Software Foundation.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

from . import _wrapper
from .exceptions import raise_if_slycot_error

import numpy as np


def mb03rd(n, A, X=None, jobx='U', sort='N', pmax=1.0, tol=0.0):
    """Ar, Xr, blsize, W = mb03rd(n, A, [X, jobx, sort, pmax, tol])

    To reduce a matrix `A` in real Schur form to a block-diagonal form
    using well-conditioned non-orthogonal similarity transformations.
    The condition numbers of the transformations used for reduction
    are roughly bounded by `pmax`*`pmax`, where `pmax` is a given value.
    The transformations are optionally postmultiplied in a given
    matrix `X`. The real Schur form is optionally ordered, so that
    clustered eigenvalues are grouped in the same block.

    Parameters
    ----------
        n : int
            The order of the matrices `A` and `X`.  `n` >= 0.
        A : (n, n) array_like
            The matrix `A` to be block-diagonalized, in real Schur form.
        X : (n, n) array_like, optional
            A given matrix `X`, for accumulation of transformations (only if
            `jobx`='U')
        jobx : {'N', 'U'}, optional
            Specifies whether or not the transformations are
            accumulated, as follows:

            := 'N': The transformations are not accumulated
            := 'U': The transformations are accumulated in `Xr` (default)

        sort : {'N', 'S', 'C', 'B'}, optional
            Specifies whether or not the diagonal blocks of the real
            Schur form are reordered, as follows:

            := 'N':  The diagonal blocks are not reordered (default);
            := 'S':  The diagonal blocks are reordered before each
                     step of reduction, so that clustered eigenvalues
                     appear in the same block;
            := 'C':  The diagonal blocks are not reordered, but the
                     "closest-neighbour" strategy is used instead of
                     the standard "closest to the mean" strategy
                     (see Notes_);
            := 'B':  The diagonal blocks are reordered before each
                     step of reduction, and the "closest-neighbour"
                     strategy is used (see Notes_).

        pmax : float, optional
            An upper bound for the infinity norm of elementary
            submatrices of the individual transformations used for
            reduction (see Notes_).  `pmax` >= 1.0
        tol : float, optional
            The tolerance to be used in the ordering of the diagonal
            blocks of the real Schur form matrix.
            If the user sets `tol` > 0, then the given value of `tol` is
            used as an absolute tolerance: a block `i` and a temporarily
            fixed block 1 (the first block of the current trailing
            submatrix to be reduced) are considered to belong to the
            same cluster if their eigenvalues satisfy

            .. math:: | \\lambda_1 - \\lambda_i | <= tol.

            If the user sets `tol` < 0, then the given value of tol is
            used as a relative tolerance: a block i and a temporarily
            fixed block 1 are considered to belong to the same cluster
            if their eigenvalues satisfy, for ``j = 1, ..., n``

            .. math:: | \\lambda_1 - \\lambda_i | <= | tol | * \\max | \\lambda_j |.

            If the user sets `tol` = 0, then an implicitly computed,
            default tolerance, defined by ``tol = SQRT( SQRT( EPS ) )``
            is used instead, as a relative tolerance, where `EPS` is
            the machine precision (see LAPACK Library routine DLAMCH).
            If `sort` = 'N' or 'C', this parameter is not referenced.

    Returns
    -------
        Ar : (n, n) ndarray
            Contains the computed block-diagonal matrix, in real Schur
            canonical form. The non-diagonal blocks are set to zero.
        Xr : (n, n) ndarray or None
            Contains the product of the given matrix `X` and the
            transformation matrix that reduced `A` to block-diagonal
            form. The transformation matrix is itself a product of
            non-orthogonal similarity transformations having elements
            with magnitude less than or equal to `pmax`.
            If `jobx` = 'N', this array is returned as None
        blsize : (n,) ndarray
            The orders of the resulting diagonal blocks of the matrix `Ar`.
        W : (n,) complex ndarray
            Contains the complex eigenvalues of the matrix `A`.

    Notes
    -----
    **Method**

    Consider first that `sort` = 'N'. Let

    ::

           ( A    A   )
           (  11   12 )
       A = (          ),
           ( 0    A   )
           (       22 )

    be the given matrix in real Schur form, where initially :math:`A_{11}` is the
    first diagonal block of dimension 1-by-1 or 2-by-2. An attempt is
    made to compute a transformation matrix `X` of the form

    ::

           ( I   P )
       X = (       )                                               (1)
           ( 0   I )

    (partitioned as `A`), so that

    ::

                ( A     0  )
        -1      (  11      )
       X  A X = (          ),
                ( 0    A   )
                (       22 )

    and the elements of `P` do not exceed the value `pmax` in magnitude.
    An adaptation of the standard method for solving Sylvester
    equations [1]_, which controls the magnitude of the individual
    elements of the computed solution [2]_, is used to obtain matrix `P`.
    When this attempt failed, an 1-by-1 (or 2-by-2) diagonal block of
    :math:`A_{22}`  , whose eigenvalue(s) is (are) the closest to the mean of those
    of :math:`A_{11}`   is selected, and moved by orthogonal similarity
    transformations in the leading position of :math:`A_{22}`  ; the moved diagonal
    block is then added to the block :math:`A_{11}`  , increasing its order by 1
    (or 2). Another attempt is made to compute a suitable
    transformation matrix X with the new definitions of the blocks :math:`A_{11}`
    and :math:`A_{22}`  . After a successful transformation matrix `X` has been
    obtained, it postmultiplies the current transformation matrix
    (if `jobx` = 'U'), and the whole procedure is repeated for the
    matrix :math:`A_{22}`.

    When `sort` = 'S', the diagonal blocks of the real Schur form are
    reordered before each step of the reduction, so that each cluster
    of eigenvalues, defined as specified in the definition of TOL,
    appears in adjacent blocks. The blocks for each cluster are merged
    together, and the procedure described above is applied to the
    larger blocks. Using the option `sort` = 'S' will usually provide
    better efficiency than the standard option (`sort` = 'N'), proposed
    in [2]_, because there could be no or few unsuccessful attempts
    to compute individual transformation matrices `X` of the form (1).
    However, the resulting dimensions of the blocks are usually
    larger; this could make subsequent calculations less efficient.

    When `sort` = 'C' or 'B', the procedure is similar to that for
    `sort` = 'N' or 'S', respectively, but the block of :math:`A_{22}` whose
    eigenvalue(s) is (are) the closest to those of :math:`A_{11}` (not to their
    mean) is selected and moved to the leading position of :math:`A_{22}`. This
    is called the "closest-neighbour" strategy.

    **Numerical Aspects**

    The algorithm usually requires :math:`\\mathcal{O}(N^3)` operations,
    but :math:`\\mathcal{O}(N^4)` are
    possible in the worst case, when all diagonal blocks in the real
    Schur form of `A` are 1-by-1, and the matrix cannot be diagonalized
    by well-conditioned transformations.

    **Further Comments**

    The individual non-orthogonal transformation matrices used in the
    reduction of `A` to a block-diagonal form have condition numbers
    of the order `pmax`*`pmax`. This does not guarantee that their product
    is well-conditioned enough. The routine can be easily modified to
    provide estimates for the condition numbers of the clusters of
    eigenvalues.

    **Contributor**

    V. Sima, Katholieke Univ. Leuven, Belgium, June 1998.
    Partly based on the RASP routine BDIAG by A. Varga, German
    Aerospace Center, DLR Oberpfaffenhofen.

    **Revisions**

    \\V. Sima, Research Institute for Informatics, Bucharest, Apr. 2003.

    References
    ----------
    .. [1] Bartels, R.H. and Stewart, G.W.  T
           Solution of the matrix equation A X + XB = C.
           Comm. A.C.M., 15, pp. 820-826, 1972.

    .. [2] Bavely, C. and Stewart, G.W.
           An Algorithm for Computing Reducing Subspaces by Block
           Diagonalization.
           SIAM J. Numer. Anal., 16, pp. 359-367, 1979.

    .. [3] Demmel, J.
           The Condition Number of Equivalence Transformations that
           Block Diagonalize Matrix Pencils.
           SIAM J. Numer. Anal., 20, pp. 599-610, 1983.

    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['jobx', 'sort', 'n', 'pmax',
                'A', 'lda' + hidden, 'X', 'ldx' + hidden,
                'nblcks', 'blsize', 'wr', 'wi', 'tol',
                'dwork' + hidden, 'info']

    if X is None:
        X = np.zeros((1, n))

    Ar, Xr, nblcks, blsize, wr, wi, info = _wrapper.mb03rd(
        jobx, sort, n, pmax, A, X, tol)

    raise_if_slycot_error(info, arg_list)
    if jobx == 'N':
        Xr = None
    else:
        Xr = Xr[:n, :n]
    Ar = Ar[:n, :n]
    W = wr + 1J*wi
    return Ar, Xr, blsize[:nblcks], W


def mb03vd(n, ilo, ihi, A):
    """ HQ, Tau = mb03vd(n, ilo, ihi, A)

    To reduce a product of p real general matrices A = A_1*A_2*...*A_p
    to upper Hessenberg form, H = H_1*H_2*...*H_p, where H_1 is
    upper Hessenberg, and H_2, ..., H_p are upper triangular, by using
    orthogonal similarity transformations on A,

            Q_1' * A_1 * Q_2 = H_1,
            Q_2' * A_2 * Q_3 = H_2,
                   ...
            Q_p' * A_p * Q_1 = H_p.

    Parameters
    ----------

    n : int
            The order of the square matrices A_1, A_2, ..., A_p.
            n >= 0.

    ilo, ihi : int
            It is assumed that all matrices A_j, j = 2, ..., p, are
            already upper triangular in rows and columns [:ilo-1] and
            [ihi:n], and A_1 is upper Hessenberg in rows and columns
            [:ilo-1] and [ihi:n], with A_1[ilo-1,ilo-2] = 0 (unless
            ilo = 1), and A_1[ihi,ihi-1] = 0 (unless ihi = n).
            If this is not the case, ilo and ihi should be set to 1
            and n, respectively.
            1 <= ilo <= max(1,n); min(ilo,n) <= ihi <= n.

    A : ndarray
            A[:n,:n,:p] must contain the matrices of factors to be reduced;
            specifically, A[:,:,j-1] must contain A_j, j = 1, ..., p.


    Returns
    -------

    HQ : ndarray
            3D array with same shape as A. The upper triangle and the first
            subdiagonal of HQ[:n,:n,0] contain the upper Hessenberg
            matrix H_1, and the elements below the first subdiagonal,
            with the first column of the array Tau represent the
            orthogonal matrix Q_1 as a product of elementary
            reflectors. See FURTHER COMMENTS.
            For j > 1, the upper triangle of HQ[:n,:n,j-1]
            contains the upper triangular matrix H_j, and the elements
            below the diagonal, with the j-th column of the array TAU
            represent the orthogonal matrix Q_j as a product of
            elementary reflectors. See FURTHER COMMENTS.

    Tau : ndarray
            2D array with shape (max(1, n-1), p).
            The leading n-1 elements in the j-th column contain the
            scalar factors of the elementary reflectors used to form
            the matrix Q_j, j = 1, ..., p. See FURTHER COMMENTS.

    Further Comments
    ----------------

    Each matrix Q_j is represented as a product of (ihi-ilo)
    elementary reflectors,

       Q_j = H_j(ilo) H_j(ilo+1) . . . H_j(ihi-1).

    Each H_j(i), i = ilo, ..., ihi-1, has the form

       H_j(i) = I - tau_j * v_j * v_j',

    where tau_j is a real scalar, and v_j is a real vector with
    v_j(1:i) = 0, v_j(i+1) = 1 and v_j(ihi+1:n) = 0; v_j(i+2:ihi)
    is stored on exit in A_j(i+2:ihi,i), and tau_j in TAU(i,j).

    The contents of A_1 are illustrated by the following example
    for n = 7, ilo = 2, and ihi = 6:

    on entry                         on exit

    ( a   a   a   a   a   a   a )    ( a   h   h   h   h   h   a )
    ( 0   a   a   a   a   a   a )    ( 0   h   h   h   h   h   a )
    ( 0   a   a   a   a   a   a )    ( 0   h   h   h   h   h   h )
    ( 0   a   a   a   a   a   a )    ( 0   v2  h   h   h   h   h )
    ( 0   a   a   a   a   a   a )    ( 0   v2  v3  h   h   h   h )
    ( 0   a   a   a   a   a   a )    ( 0   v2  v3  v4  h   h   h )
    ( 0   0   0   0   0   0   a )    ( 0   0   0   0   0   0   a )

    where a denotes an element of the original matrix A_1, h denotes
    a modified element of the upper Hessenberg matrix H_1, and vi
    denotes an element of the vector defining H_1(i).

    The contents of A_j, j > 1, are illustrated by the following
    example for n = 7, ilo = 2, and ihi = 6:

    on entry                         on exit

    ( a   a   a   a   a   a   a )    ( a   h   h   h   h   h   a )
    ( 0   a   a   a   a   a   a )    ( 0   h   h   h   h   h   h )
    ( 0   a   a   a   a   a   a )    ( 0   v2  h   h   h   h   h )
    ( 0   a   a   a   a   a   a )    ( 0   v2  v3  h   h   h   h )
    ( 0   a   a   a   a   a   a )    ( 0   v2  v3  v4  h   h   h )
    ( 0   a   a   a   a   a   a )    ( 0   v2  v3  v4  v5  h   h )
    ( 0   0   0   0   0   0   a )    ( 0   0   0   0   0   0   a )

    where a denotes an element of the original matrix A_j, h denotes
    a modified element of the upper triangular matrix H_j, and vi
    denotes an element of the vector defining H_j(i). (The element
    (1,2) in A_p is also unchanged for this example.)

    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['n', 'p' + hidden,
                'ilo', 'ihi', 'a',
                'lda1' + hidden, 'lda2' + hidden, 'tau',
                'ldtau' + hidden, 'dwork' + hidden, 'info']

    HQ, Tau, info = _wrapper.mb03vd(n, ilo, ihi, A)

    raise_if_slycot_error(info, arg_list)
    return (HQ, Tau)


def mb03vy(n, ilo, ihi, A, Tau, ldwork=None):
    """ Q = mb03vy(n, ilo, ihi, A, Tau, [ldwork])

    To generate the real orthogonal matrices Q_1, Q_2, ..., Q_p,
    which are defined as the product of ihi-ilo elementary reflectors
    of order n, as returned by SLICOT Library routine MB03VD:

        Q_j = H_j(ilo) H_j(ilo+1) . . . H_j(ihi-1).

    Parameters
    ----------

    n : int
            The order of the matrices Q_1, Q_2, ..., Q_p.  N >= 0.

    ilo, ihi : int
            The values of the indices ilo and ihi, respectively, used
            in the previous call of the SLICOT Library routine MB03VD.
            1 <= ilo <= max(1,n); min(ilo,n) <= ihi <= n.

    A : ndarray
            A[:n,:n,j-1] must contain the vectors which define the
            elementary reflectors used for reducing A_j, as returned
            by SLICOT Library routine MB03VD, j = 1, ..., p.

    Tau : ndarray
            The leading N-1 elements in the j-th column must contain
            the scalar factors of the elementary reflectors used to
            form the matrix Q_j, as returned by SLICOT Library routine
            MB03VD.

    ldwork : int, optional
            The length of the internal array DWORK.  ldwork >= max(1, n).
            For optimum performance ldwork should be larger.


    Returns
    -------

    Q : ndarray
            3D array with same shape as A. Q[:n,:n,j-1] contains the
            N-by-N orthogonal matrix Q_j, j = 1, ..., p.

    """

    hidden = ' (hidden by the wrapper)'
    arg_list = ['n', 'p' + hidden,
                'ilo', 'ihi', 'a',
                'lda1' + hidden, 'lda2' + hidden, 'tau',
                'ldtau' + hidden, 'dwork' + hidden, 'ldwork', 'info' + hidden]

    if not ldwork:
        ldwork = max(1, 2 * n)

    Q, info = _wrapper.mb03vy(n, ilo, ihi, A, Tau, ldwork)

    raise_if_slycot_error(info, arg_list)

    return Q


def mb03wd(job, compz, n, ilo, ihi, iloz, ihiz, H, Q, ldwork=None):
    """ T, Z, Wr = mb03wd(job, compz, n, ilo, ihi, iloz, ihiz, H, Q, [ldwork])

    To compute the Schur decomposition and the eigenvalues of a
    product of matrices, H = H_1*H_2*...*H_p, with H_1 an upper
    Hessenberg matrix and H_2, ..., H_p upper triangular matrices,
    without evaluating the product. Specifically, the matrices Z_i
    are computed, such that

    ::

            Z_1' * H_1 * Z_2 = T_1,
            Z_2' * H_2 * Z_3 = T_2,
                   ...
            Z_p' * H_p * Z_1 = T_p,

    where T_1 is in real Schur form, and T_2, ..., T_p are upper
    triangular.

    The routine works primarily with the Hessenberg and triangular
    submatrices in rows and columns ilo to ihi, but optionally applies
    the transformations to all the rows and columns of the matrices
    H_i, i = 1,...,p. The transformations can be optionally
    accumulated.

    Parameters
    ----------
    job : {'E', 'S'}
            Indicates whether the user wishes to compute the full
            Schur form or the eigenvalues only, as follows:
            = 'E':  Compute the eigenvalues only;
            = 'S':  Compute the factors T_1, ..., T_p of the full
                    Schur form, T = T_1*T_2*...*T_p.
   compz : {'N', 'I', 'V'}
            Indicates whether or not the user wishes to accumulate
            the matrices Z_1, ..., Z_p, as follows:
            = 'N':  The matrices Z_1, ..., Z_p are not required;
            = 'I':  Z_i is initialized to the unit matrix and the
                    orthogonal transformation matrix Z_i is returned,
                    i = 1, ..., p;
            = 'V':  Z_i must contain an orthogonal matrix Q_i on
                    entry, and the product Q_i*Z_i is returned,
                    i = 1, ..., p.
    n : int
            The order of the matrix H.  n >= 0
    ilo, ihi : int
            It is assumed that all matrices H_j, j = 2, ..., p, are
            already upper triangular in rows and columns [:ilo-1] and
            [ihi:n], and H_1 is upper quasi-triangular in rows and
            columns [:ilo-1] and [ihi:n], with H_1[ilo-1,ilo-2] = 0
            (unless ilo = 1), and H_1[ihi,ihi-1] = 0 (unless ihi = n).
            The routine works primarily with the Hessenberg submatrix
            in rows and columns ilo to ihi, but applies the
            transformations to all the rows and columns of the
            matrices H_i, i = 1,...,p, if JOB = 'S'.
            1 <= ilo <= max(1,n); min(ilo,n) <= ihi <= n.
    iloz, ihiz : int
            Specify the rows of Z to which the transformations must be
            applied if compz = 'I' or compz = 'V'.
            1 <= iloz <= ilo; ihi <= ihiz <= n.
    H : array_like
            H[:n,:n,0] must contain the upper Hessenberg matrix H_1 and
            H[:n,:n,j-1] for j > 1 must contain the upper triangular matrix
            H_j, j = 2, ..., p.
    Q : array_like
            If compz = 'V', Q[:n,:n,:p] must contain the current matrix Q of
            transformations accumulated by SLICOT Library routine
            MB03VY.
            If compz = 'I', Q is ignored
    ldwork : int, optional
            The length of the cache array. The default value is
            ihi-ilo+p-1

    Returns
    -------
    T : ndarray
            3D array with the same shape as H.
            If JOB = 'S', T[:n,:n,0] is upper quasi-triangular in rows
            and columns [ilo-1:ihi], with any 2-by-2 diagonal blocks
            corresponding to a pair of complex conjugated eigenvalues, and
            T[:n,:n,j-1] for j > 1 contains the resulting upper
            triangular matrix T_j.
            If job = 'E', T is None
    Z : ndarray
            3D array with the same shape as Q.
            If compz = 'V', or compz = 'I', the leading
            N-by-N-by-P part of this array contains the transformation
            matrices which produced the Schur form; the
            transformations are applied only to the submatrices
            Z[iloz-1:ihiz,ilo-1:ihi,j-1], j = 1, ..., p.
            If compz = 'N', Z is None
    W : (n,) complex ndarray
            The computed eigenvalues ilo to ihi. If two eigenvalues
            are computed as a complex conjugate pair, they are stored
            in consecutive elements of W say the i-th and
            (i+1)th, with imag(W][i]) > 0 and imag(W[i+1]) < 0.
            If JOB = 'S', the eigenvalues are stored in the same order
            as on the diagonal of the Schur form returned in H.

    Warns
    -----
    SlycotResultWarning
        :info > 0:
            failed to compute all the eigenvalues {ilo} to {ihi}
            in a total of 30*({ihi}-{ilo}+1) iterations
            the elements Wr[{info}:{ihi}] contains those
            eigenvalues which have been successfully computed.
    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['job', 'compz', 'n', 'p' + hidden,
                'ilo', 'ihi', 'iloz', 'ihiz',
                'h', 'ldh1' + hidden, 'ldh2' + hidden,
                'z', 'ldz1' + hidden, 'ldz2' + hidden,
                'wr', 'wi',
                'dwork' + hidden, 'ldwork', 'info' + hidden]

    if not ldwork:
        p = H.shape[2]
        ldwork = max(1, ihi - ilo + p - 1)

    T, Z, Wr, Wi, info = _wrapper.mb03wd(
            job, compz, n, ilo, ihi, iloz, ihiz, H, Q, ldwork)
    raise_if_slycot_error(info, arg_list, mb03rd.__doc__, locals())

    if job == 'E':
        T = None
    if compz == 'N':
        Z = None

    W = Wr + Wi*1J
    return (T, Z, W)


def mb05md(a, delta, balanc='N'):
    """Ar, Vr, Yr, w = mb05md(a, delta, balanc='N')

    Matrix exponential for a real non-defective matrix

    To compute ``exp(A*delta)`` where `A` is a real N-by-N non-defective
    matrix with real or complex eigenvalues and delta is a scalar
    value. The routine also returns the eigenvalues and eigenvectors
    of `A` as well as (if all eigenvalues are real) the matrix product
    ``exp(Lambda*delta)`` times the inverse of the eigenvector matrix of
    `A`, where `Lambda` is the diagonal matrix of eigenvalues.
    Optionally, the routine computes a balancing transformation to
    improve the conditioning of the eigenvalues and eigenvectors.

    Parameters
    ----------
    A : (n, n) array_like
        Square matrix
    delta : float
        The scalar value delta of the problem.
    balanc : {'N', 'S'}, optional
        Indicates how the input matrix should be diagonally scaled
        to improve the conditioning of its eigenvalues as follows:

        := 'N':  Do not diagonally scale;
        := 'S':  Diagonally scale the matrix, i.e. replace `A` by
                 ``D*A*D**(-1)``, where `D` is a diagonal matrix chosen
                 to make the rows and columns of A more equal in
                 norm. Do not permute.

    Returns
    -------
    Ar : (n, n) ndarray
        Contains the solution matrix ``exp(A*delta)``
    Vr : (n, n) ndarray
        Contains the eigenvector matrix for `A`.  If the `k`-th
        eigenvalue is real the `k`-th column of the eigenvector
        matrix holds the eigenvector corresponding to the `k`-th
        eigenvalue.  Otherwise, the `k`-th and `(k+1)`-th eigenvalues
        form a complex conjugate pair and the k-th and `(k+1)`-th
        columns of the eigenvector matrix hold the real and
        imaginary parts of the eigenvectors corresponding to these
        eigenvalues as follows.  If `p` and `q` denote the `k`-th and
        `(k+1)`-th columns of the eigenvector matrix, respectively,
        then the eigenvector corresponding to the complex
        eigenvalue with positive (negative) imaginary value is
        given by
          ``p + q*j (p - q*j), where j^2  = -1.``
    Yr : (n, n) ndarray
        contains an intermediate result for computing the matrix
        exponential.  Specifically, ``exp(A*delta)`` is obtained as the
        product ``V*Y``, where `V` is the matrix stored in the leading
        `n`-by-`n` part of the array `V`. If all eigenvalues of `A` are
        real, then the leading `n`-by-`n` part of this array contains
        the matrix product ``exp(Lambda*delta)`` times the inverse of
        the (right) eigenvector matrix of `A`, where `Lambda` is the
        diagonal matrix of eigenvalues.
    w : (n, ) real or complex ndarray
        Contains the eigenvalues of the matrix `A`. The eigenvalues
        are unordered except that complex conjugate pairs of values
        appear consecutively with the eigenvalue having positive
        imaginary part first.

    Warns
    ------
    SlycotResultWarning
        :0 < info <=n:
            the QR algorithm failed to compute all
            the eigenvalues; no eigenvectors have been computed;
            w[{info}:{n}] contains eigenvalues which have converged;
        :info == n+1:
            The inverse of the eigenvector matrix could not
            be formed due to an attempt to divide by zero, i.e.,
            the eigenvector matrix is singular;
        :info == n+2:
            Matrix A is defective, possibly due to rounding errors.
    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['balanc', 'n', 'delta', 'a', 'lda'+hidden, 'v', 'ldv'+hidden,
                'y', 'ldy'+hidden, 'valr', 'vali',
                'iwork'+hidden, 'dwork'+hidden, 'ldwork'+hidden,
                'info'+hidden]
    n = min(a.shape)
    (Ar, Vr, Yr, VALr, VALi, INFO) = _wrapper.mb05md(balanc=balanc,
                                                     n=n,
                                                     delta=delta,
                                                     a=a)
    raise_if_slycot_error(INFO, arg_list, mb05md.__doc__, locals())

    if not all(VALi == 0):
        w = VALr + 1J*VALi
    else:
        w = VALr
    return (Ar, Vr, Yr, w)



def mb05nd(a, delta, tol=1e-7):
    """F, H = mb05nd(n, a, delta, tol=1e-7)

    To compute

    ::

     (a)    F(delta) =  exp(A*delta) and
     (b)    H(delta) =  Int[F(s) ds] from s = 0 to s = delta,

    where `A` is a real`n`-by-`n` matrix and `delta` is a scalar value.

    Parameters
    ----------
    A : (n, n) array_like
        Square matrix
    delta : float
        The scalar value delta of the problem.
    tol : float
        Tolerance. A good value is sqrt(eps)

    Returns
    -------
    F : (n, n) ndarray
        exp(A*delta)
    H : (n, n) ndarray
        Int[F(s) ds] from s = 0 to s = delta,

    Raises
    ------
    SlycotArithmeticError
        :1 < info <=n:
            the ({info},{info}) element of the denominator of
            the Pade approximation is zero, so the denominator
            is exactly singular;
        :info == n+1:
            ``DELTA = (delta * frobenius norm of matrix A)`` is
            probably too large to permit meaningful computation.
            That is, {delta} > SQRT(BIG), where BIG is a
            representable number near the overflow threshold of
            the machine (see LAPACK Library Routine DLAMCH).
    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['n', 'delta', 'a', 'lda'+hidden, 'ex', 'ldex'+hidden,
                'exint', 'ldexin'+hidden, 'tol', 'iwork'+hidden,
                'dwork'+hidden, 'ldwork'+hidden]
    n = min(a.shape)
    out = _wrapper.mb05nd(n=n, delta=delta, a=a, tol=tol)
    raise_if_slycot_error(out[-1], arg_list, mb05nd.__doc__, locals())
    return out[:-1]



def mc01td(dico, dp, p):
    """dp, stable, nz = mc01td(dico, dp, p)

    To determine whether or not a given polynomial P(x) with real
    coefficients is stable, either in the continuous-time or discrete-
    time case.

    A polynomial is said to be stable in the continuous-time case
    if all its zeros lie in the left half-plane, and stable in the
    discrete-time case if all its zeros lie inside the unit circle.


    Parameters
    ----------
    dico : {'C', 'D'}
        Indicates whether the stability test to be applied to `P(x)` is in
        the continuous-time or discrete-time case as follows::

        = 'C':  continuous-time case;
        = 'D':  discrete-time case.

    dp : int
        The degree of the polynomial `P(x)`.  ``dp >= 0``.
    p : (dp+1, ) array_like
        This array must contain the coefficients of `P(x)` in increasing
        powers of `x`.

    Returns
    -------
    dp : int
        If ``P(dp+1) = 0.0`` on entry, then `dp` contains the index of the
        highest power of `x` for which ``P(dp+1) <> 0.0``.
    stable : int
        Equal to 1 if `P(x)` is stable, 0 otherwise.
    nz : int
        The number of unstable zeros.

    Warns
    -----
    SlycotResultWarning
        :info == 1:
            Entry ``P(x)`` is the zero polynomial.
        :info == 2 and dico == 'C':
            The polynomial ``P(x)`` is most probably unstable,
            although it may be stable with one or more zeros
            very close to the imaginary axis.
            The number of unstable zeros (NZ) is not determined.
        :info == 2 and dico == 'D':
            The polynomial ``P(x)`` is most probably unstable,
            although it may be stable with one or more zeros
            very close to the the unit circle.
            The number of unstable zeros (NZ) is not determined.
        :iwarn > 0:
            The degree of the polynomial ``P(x)`` has been
            reduced to ``(DB - {iwarn})`` because
            ``P(DB+1-j) = 0.0`` on entry
            for ``j = 0, 1,..., k-1`` and ``P(DB+1-k) <> 0.0``.
    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['dico', 'dp', 'P', 'stable', 'nz', 'DWORK' + hidden,
                'IWARN', 'INFO']
    (dp_out, stable_log, nz, iwarn, info) = _wrapper.mc01td(dico, dp, p)
    raise_if_slycot_error([iwarn, info], arg_list, mc01td.__doc__, locals())
    ftrue, ffalse = _wrapper.ftruefalse()
    stable = 1 if stable_log == ftrue else 0
    return (dp_out, stable, nz)



#
# test_td04ad.py - test suite for tf -> ss conversion
# RvP, 04 Jun 2018

from __future__ import print_function, division

import unittest
from slycot import transform
import numpy as np


class TestTf2SS(unittest.TestCase):

    def test_td04ad_c(self):
        """td04ad: Convert with 'C' option"""

        # for octave:
        """
        num = { [0.0,  0.0, 1.0 ], [ 1.0, 0.0 ];
                [3.0, -1.0, 1.0 ], [ 0.0, 1.0 ];
                [0.0, 0.0, 1.0],   [ 0.0, 2.0 ] };
        den = { [1.0,  0.4, 3.0],  [ 1.0, 1.0 ];
                [1.0,  0.4, 3.0],  [ 1.0, 1.0 ];
                [1.0,  0.4, 3.0],  [ 1.0, 1.0 ]};
        """

        m = 2
        p = 3
        d = 3
        num = np.array([
            [ [0.0,  0.0, 1.0], [1.0, 0.0, 0.0] ],
            [ [3.0, -1.0, 1.0], [0.0, 1.0, 0.0] ],
            [ [0.0,  0.0, 1.0], [0.0, 2.0, 0.0] ] ])

        numc = np.zeros((max(1, m, p), max(1, m, p), d), dtype=float)
        numc[:p,:m,:] = num
        denc = np.array(
            [ [1.0,  0.4, 3.0],  [ 1.0, 1.0, 0.0 ] ])
        indc = np.array(
            [ 2, 1 ], dtype=int)

        nref = 3
        Aref = np.array([ [-1,    0,    0],
                          [ 0, -0.4, -0.3],
                          [ 0,   10,    0] ])
        Bref = np.array([ [0, -1],
                          [1,  0],
                          [0,  0] ])
        Cref = np.array([ [1,     0,  0.1],
                          [-1, -2.2, -0.8],
                          [-2,    0,  0.1] ])
        Dref = np.array([ [0, 1],
                          [3, 0],
                          [0, 0] ])

        nr, A, B, C, D = transform.td04ad('C', m, p, indc, denc, numc)
        #print('A=\n', A, '\nB=\n', B, '\nC=\n', C, '\nD=\n', D)
        np.testing.assert_equal(nref, nr)
        # the returned state space representation is not guaranteed to
        # be of one form for all architectures, so we transform back
        # to tf and check for equality then
        _, _, _, _, _, dcoeff, ucoeff = transform.tb04ad(
                nr, m, p, A, B, C, D)
        _, _, _, _, _, dcoeffref, ucoeffref = transform.tb04ad(
                nref, m, p, Aref, Bref, Cref, Dref)
        np.testing.assert_array_almost_equal(dcoeff,dcoeffref)
        np.testing.assert_array_almost_equal(ucoeff,ucoeffref)

    def test_td04ad_r(self):
        """td04ad: Convert with 'R' option

        example program from
        http://slicot.org/objects/software/shared/doc/TD04AD.html
        """

        m = 2
        p = 2
        rowcol = 'R'
        index = [3, 3]
        dcoeff = np.array([ [1.0, 6.0, 11.0, 6.0], [1.0, 6.0, 11.0, 6.0] ])

        ucoeff = np.array([ [[1.0, 6.0, 12.0, 7.0], [0.0, 1.0,  4.0,  3.0]],
                            [[0.0, 0.0, 1.0,  1.0], [1.0, 8.0, 20.0, 15.0]] ])

        nref = 3

        Aref = np.array([ [ 0.5000,  -0.8028,   0.9387],
                          [ 4.4047,  -2.3380,   2.5076],
                          [-5.5541,   1.6872,  -4.1620] ])
        Bref = np.array([ [-0.2000,  -1.2500],
                          [ 0.0000,  -0.6097],
                          [ 0.0000,   2.2217] ])
        Cref = np.array([ [0.0000,  -0.8679,   0.2119],
                          [0.0000,   0.0000,   0.9002] ])
        Dref = np.array([ [1.0000,   0.0000],
                          [0.0000,   1.0000] ])

        nr, A, B, C, D = transform.td04ad(rowcol, m, p, index, dcoeff, ucoeff)
        #print('A=\n', A, '\nB=\n', B, '\nC=\n', C, '\nD=\n', D)
        np.testing.assert_equal(nref, nr)
        # order of states is not guaranteed, so we reorder the reference
        rindex = np.flip(np.argsort(np.diag(A)))
        Arref = Aref[rindex, :][:, rindex]
        Brref = Bref[rindex, :]
        Crref = Cref[:, rindex]
        Drref = Dref
        np.testing.assert_array_almost_equal(A, Arref,decimal=4)
        np.testing.assert_array_almost_equal(B, Brref,decimal=4)
        np.testing.assert_array_almost_equal(C, Crref,decimal=4)
        np.testing.assert_array_almost_equal(D, Drref,decimal=4)


    def test_staticgain(self):
        """td04ad: Convert a transferfunction to SS with only static gain"""

        # 2 inputs, 3 outputs? columns share a denominator
        num = np.array([ [ [1.0], [2.0] ],
                         [ [0.2], [4.3] ],
                         [ [1.2], [3.2] ] ])
        p, m, d = num.shape
        numc = np.zeros((max(1, m, p), max(1, m, p), d), dtype=float)
        numc[:p,:m,:] = num

        # denc, columns share a common denominator
        denc = np.array([ [ 1.0], [0.5] ])
        Dc = (num / denc).reshape((3,2))
        idxc = np.zeros((2,), dtype=int)

        # denr, rows share a common denominator
        denr = np.array([ [1.0], [0.5], [3.0] ])
        idxr = np.zeros((3,), dtype=int)
        Dr = (num / denr[:, np.newaxis]).reshape((3,2))

        # fails with:
        # On entry to TB01XD parameter number  5 had an illegal value

        n, A, B, C, D = transform.td04ad('C', 2, 3, idxc, denc, numc)
        #print('A=\n', A, '\nB=\n', B, '\nC=\n', C, '\nD=\n', D)
        self.assertEqual(A.shape, (0,0))
        self.assertEqual(B.shape, (0,2))
        self.assertEqual(C.shape, (3,0))
        np.testing.assert_array_almost_equal(D, Dc)

        n, A, B, C, D = transform.td04ad('R', 2, 3, idxr, denr, num)
        #print('A=\n', A, '\nB=\n', B, '\nC=\n', C, '\nD=\n', D)
        self.assertEqual(A.shape, (0,0))
        self.assertEqual(B.shape, (0,2))
        self.assertEqual(C.shape, (3,0))
        np.testing.assert_array_almost_equal(D, Dr)

    def test_td04ad_static(self):
        """Regression: td04ad (TFM -> SS transformation) for static TFM"""
        from itertools import product
        for nout, nin, rc in product(range(1, 6), range(1, 6), ['R', 'C']):
            Dref = np.zeros((nout, nin))
            if rc == 'R':
                num = np.reshape(np.arange(nout * nin), (nout, nin, 1))
                den = np.reshape(np.arange(1, 1 + nout), (nout, 1))
                index = np.repeat(0, nout)
                Dref = num[:nout, :nin, 0] / np.broadcast_to(den, (nout, nin))
            else:
                maxn = max(nout, nin)
                num = np.zeros((maxn, maxn, 1))
                num[:nout, :nin, 0] = np.reshape(
                        np.arange(nout * nin), (nout, nin))
                den = np.reshape(np.arange(1, 1 + nin), (nin, 1))
                index = np.repeat(0, nin)
                Dref = num[:nout, :nin, 0] / np.broadcast_to(den.T, (nout, nin))
            nr, A, B, C, D = transform.td04ad(rc, nin, nout, index, den, num)
            np.testing.assert_equal(nr, 0)
            for M in [A, B, C]:
                np.testing.assert_equal(M, np.zeros_like(M))
            np.testing.assert_almost_equal(D, Dref)

    def test_mixfeedthrough(self):
        """Test case popping up from control testing

        a mix of feedthrough and dynamics. The problem from the control
        package was somewhere else
        """
        num = np.array([ [ [ 0.0,  0.0 ], [ 0.0, -0.2 ] ],
                         [ [ -0.1,  0.0 ], [ 0.0,  0.0 ] ] ])
        p, m, d = num.shape
        numc = np.zeros((max(1, m, p), max(1, m, p), d), dtype=float)
        numc[:p,:m,:] = num
        denc = np.array([[1.0, 1.1],
                         [1.0, 0.0]])
        idxc = np.array([1, 0])
        n, A, B, C, D = transform.td04ad('C', 2, 2, idxc, denc, numc)
        np.testing.assert_array_almost_equal(D, np.array([[0,  0],[-0.1, 0]]))

    def test_toandfrom(self):
        A = np.array([[-3.0]])
        B = np.array([[0.1, 0.0]])
        C = np.array([[1.0],
                      [0.0]])
        D = np.array([[0.0, 0.0],
                      [0.0, 1.0]])

        tfout = transform.tb04ad(1, 2, 2, A, B, C, D)

        num = tfout[6]
        den = tfout[5]
        idxc = np.array([1, 0])
        n, At, Bt, Ct, Dt = transform.td04ad('R', 2, 2, idxc, den, num)
        np.testing.assert_array_almost_equal(D, Dt)
        np.testing.assert_array_almost_equal(A, At)

    def test_tfm2ss_6(self):
        """Python version of Fortran test program from
        -- Bug in TD04AD when ROWCOL='C' #6
           This bug was fixed in PR #27"""
        m = 1
        p = 1
        index = np.array([0])
        dcoeff = np.array([[0.5]])
        ucoeff = np.array([[[32]]])
        n, A, B, C, D = transform.td04ad('R', m, p, index, dcoeff, ucoeff)
        self.assertEqual(n, 0)
        np.testing.assert_array_almost_equal(D, np.array([[64]]))
        n, A, B, C, D = transform.td04ad('C', m, p, index, dcoeff, ucoeff)
        self.assertEqual(n, 0)
        np.testing.assert_array_almost_equal(D, np.array([[64]]))


if __name__ == "__main__":
    unittest.main()



#
# test_sg02ad.py - test suite for ricatti equation solving
# RvP, 19 Jun 2017
from __future__ import print_function

import unittest
from slycot import synthesis
import numpy as np

from numpy.testing import assert_almost_equal


class test_sg02ad(unittest.TestCase):

    def test_sg02ad_case1(self):
        n = 3
        m = 1
        # from a discussion here:
        # https://github.com/scipy/scipy/issues/2251
        A = np.array([[ 0.63399379,  0.54906824,  0.76253406],
                      [ 0.5404729 ,  0.53745766,  0.08731853],
                      [ 0.27524045,  0.84922129,  0.4681622 ]])
        B = np.array([[ 0.96861695],
                      [ 0.05532739],
                      [ 0.78934047]])
        Q = np.eye(3)
        E = np.eye(3)
        R = np.ones((1,1), dtype=float)
        S = np.array([[-2.67522766, -5.39447418,  2.19128542],
                      [-1.94918951, -3.15480639,  5.24379117],
                      [ 4.29133973,  8.10585767, -5.88895897]])
        L = np.array(np.zeros((3,1)))
        rcondu, X, alphar, alphai, beta, S, T, U, iwarn = \
            synthesis.sg02ad('D', 'B', 'N', 'U', 'Z', 'N', 'S', 'R',
                             n, m, 1,
                             A, E, B, Q, R, L)
        LATXB = L + A.T.dot(X).dot(B)
        assert_almost_equal(
            A.T.dot(X).dot(A) -
            E.T.dot(X).dot(E) -
            LATXB.dot(np.linalg.solve(R+B.T.dot(X).dot(B), LATXB.T)) + Q,
            np.zeros((n, n)))


if __name__ == "__main__":
    unittest.main()



# ===================================================
# ag08bd tests

import unittest
from slycot import analysis
import numpy as np

from numpy.testing import assert_almost_equal, assert_equal

# test1 input parameters

test1_l = 9
test1_n = 9
test1_m = 3
test1_p = 3
test1_tol = 1.0e-7
test1_equil = 'N'

test1_A = np.eye(9, dtype=int)

test1_E = np.array([[0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [1, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 1, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 1, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 1, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 1, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 1, 0]])

test1_B = np.array([[-1,  0,  0],
                    [ 0,  0,  0],
                    [ 0,  0,  0],
                    [ 0, -1,  0],
                    [ 0,  0,  0],
                    [ 0,  0,  0],
                    [ 0,  0, -1],
                    [ 0,  0,  0],
                    [ 0,  0,  0]])

test1_C = np.array([[ 0,  1,  1,  0,  3,  4,  0,  0,  2],
                    [ 0,  1,  0,  0,  4,  0,  0,  2,  0],
                    [ 0,  0,  1,  0, -1,  4,  0, -2,  2]])

test1_D = np.array([[ 1,  2, -2],
                    [ 0, -1, -2],
                    [ 0,  0,  0]])


class test_tg01fd(unittest.TestCase):
    """ Verify ag08bd with input parameters according to example in documentation """

    def test1_ag08bd(self):
        """test [A-lambda*E]

        B,C,D must have correct dimensions according to l,n,m and p, but cannot
        have zero length in any dimenstion. Then the wrapper will complain.
        The length is then set to one.
        """

        Af,Ef,nrank,niz,infz,kronr,infe,kronl = analysis.ag08bd(l=test1_l,n=test1_n,m=0,p=0,A=test1_A,E=test1_E,B=np.zeros((test1_l,1)),C=np.zeros((1,test1_n)),D=np.zeros((1,1)),equil=test1_equil, tol=test1_tol)

        assert_equal(Af, np.zeros((0,0)))
        assert_equal(Ef, np.zeros((0,0)))
        assert_equal(nrank, 9)
        assert_equal(niz, 6)
        assert_equal(infz, [0,3])
        assert_equal(kronr, [])
        assert_equal(infe, [3,3,3])
        assert_equal(kronl, [])

    def test2_ag08bd(self):
        """test [A-lambda*E;C]

        B,D must have correct dimensions as before
        """

        Af,Ef,nrank,niz,infz,kronr,infe,kronl = analysis.ag08bd(l=test1_l,n=test1_n,m=0,p=test1_p,A=test1_A,E=test1_E,B=np.zeros((test1_l,1)),C=test1_C,D=np.zeros((test1_p,1)),equil=test1_equil, tol=test1_tol)

        assert_equal(Af, np.zeros((0,0)))
        assert_equal(Ef, np.zeros((0,0)))
        assert_equal(nrank, 9)
        assert_equal(niz, 4)
        assert_equal(infz, [0,2])
        assert_equal(kronr, [])
        assert_equal(infe, [1,3,3])
        assert_equal(kronl, [0,1,1])

    def test3_ag08bd(self):
        """test [A-lambda*E,B]

        C,D must have correct dimensions as before
        """

        Af,Ef,nrank,niz,infz,kronr,infe,kronl = analysis.ag08bd(l=test1_l,n=test1_n,m=test1_m,p=0,A=test1_A,E=test1_E,B=test1_B,C=np.zeros((1,test1_n)),D=np.zeros((1,test1_m)),equil=test1_equil, tol=test1_tol)

        assert_equal(Af, np.zeros((0,0)))
        assert_equal(Ef, np.zeros((0,0)))
        assert_equal(nrank, 9)
        assert_equal(niz, 0)
        assert_equal(infz, [])
        assert_equal(kronr, [2,2,2])
        assert_equal(infe, [1,1,1])
        assert_equal(kronl, [])

    def test4_ag08bd(self):
        """test [A-lambda*E,B;C,D]"""

        Af,Ef,nrank,niz,infz,kronr,infe,kronl = analysis.ag08bd(l=test1_l,n=test1_n,m=test1_m,p=test1_p,A=test1_A,E=test1_E,B=test1_B,C=test1_C,D=test1_D,equil=test1_equil, tol=test1_tol)

        assert_almost_equal(Af, [[0.77045021]])
        assert_almost_equal(Ef, [[0.77045021]])
        assert_equal(nrank, 11)
        assert_equal(niz, 2)
        assert_equal(infz, [0,1])
        assert_equal(kronr, [2])
        assert_equal(infe, [1,1,1,1,3])
        assert_equal(kronl, [1])


if __name__ == "__main__":
    unittest.main()



#
#       transform.py
#
#       Copyright 2010-2011 Enrico Avventi <avventi@kth.se>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License version 2 as
#       published by the Free Software Foundation.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

from . import _wrapper
from .exceptions import raise_if_slycot_error, SlycotParameterError

import numpy as _np

def tb01id(n,m,p,maxred,a,b,c,job='A'):
    """ s_norm,A,B,C,scale = tb01id(n,m,p,maxred,A,B,C,[job])

    To reduce the 1-norm of a system matrix

          S =  ( A  B )
               ( C  0 )

    corresponding to the triple (A,B,C), by balancing. This involves
    a diagonal similarity transformation inv(D)*A*D applied
    iteratively to A to make the rows and columns of
                        -1
               diag(D,I)  * S * diag(D,I)

    as close in norm as possible.

    The balancing can be performed optionally on the following
    particular system matrices

           S = A,    S = ( A  B )    or    S = ( A )
                                               ( C )

    Required arguments:
        n : input int
            The order of the matrix A, the number of rows of matrix B and
            the number of columns of matrix C. It represents the dimension of
            the state vector.  n > 0.
        m : input int
            The number of columns of matrix B. It represents the dimension of
            the input vector.  m > 0.
        p : input int
            The number of rows of matrix C. It represents the dimension of
            the output vector.  p > 0.
        maxred : input float
            The maximum allowed reduction in the 1-norm of S  (in an iteration)
            if zero rows or columns are encountered.
            If maxred > 0.0, maxred must be larger than one (to enable the norm
            reduction).
            If maxred <= 0.0, then the value 10.0 for maxred is used.
        A : input rank-2 array('d') with bounds (n,n)
            The leading n-by-n part of this array must contain the system state
            matrix A.
        B : input rank-2 array('d') with bounds (n,m)
            The leading n-by-m part of this array must contain the system input
            matrix B.
        C : input rank-2 array('d') with bounds (p,n)
            The leading p-by-n part of this array must contain the system output
            matrix C.
    Optional arguments:
        job := 'A' input string(len=1)
            Indicates which matrices are involved in balancing, as follows:
            = 'A':  All matrices are involved in balancing;
            = 'B':  B and A matrices are involved in balancing;
            = 'C':  C and A matrices are involved in balancing;
            = 'N':  B and C matrices are not involved in balancing.
    Return objects:
        s_norm : float
            The 1-norm of the given matrix S is non-zero, the ratio between
            the 1-norm of the given matrix and the 1-norm of the balanced matrix.
        A : rank-2 array('d') with bounds (n,n)
            The leading n-by-n part of this array contains the balanced matrix
            inv(D)*A*D.
        B : rank-2 array('d') with bounds (n,m)
            The leading n-by-m part of this array contains the balanced matrix
            inv(D)*B.
        C : rank-2 array('d') with bounds (p,n)
            The leading p-by-n part of this array contains the balanced matrix C*D.
        scale : rank-1 array('d') with bounds (n)
            The scaling factors applied to S.  If D(j) is the scaling factor
            applied to row and column j, then scale(j) = D(j), for j = 1,...,n.
    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['job', 'N', 'M', 'P', 'maxred', 'A', 'LDA'+hidden, 'B',
        'LDB'+hidden, 'C', 'LDC'+hidden, 'scale', 'INFO'+hidden]
    out = _wrapper.tb01id(n,m,p,maxred,a,b,c,job=job)
    raise_if_slycot_error(out[-1], arg_list)
    return out[:-1]

def tb03ad(n,m,p,A,B,C,D,leri,equil='N',tol=0.0,ldwork=None):
    """ A_min,b_min,C_min,nr,index,pcoeff,qcoeff,vcoeff = tb03ad_l(n,m,p,A,B,C,D,leri,[equil,tol,ldwork])

    To find a relatively prime left or right polynomial matrix representation
    with the same transfer matrix as that of a given state-space representation,
    i.e. if leri = 'L'

        inv(P(s))*Q(s) = C*inv(s*I-A)*B + D

    or, if leri = 'R'

        Q(s)*inv(P(s)) = C*inv(s*I-A)*B + D.

    Additionally a minimal realization (A_min,B_min,C_min) of the original
    system (A,B,C) is returned.

    Required arguments:
        n : input int
            The order of the state-space representation, i.e. the order of
            the original state dynamics matrix A.  n > 0.
        m : input int
            The number of system inputs.  m > 0.
        p : input int
            The number of system outputs.  p > 0.
        A : input rank-2 array('d') with bounds (n,n)
            The leading n-by-n part of this array must contain the original
            state dynamics matrix A.
        B : input rank-2 array('d') with bounds (n,max(m,p))
            The leading n-by-m part of this array must contain the original
            input/state matrix B; the remainder of the leading n-by-max(m,p)
            part is used as internal workspace.
        C : input rank-2 array('d') with bounds (max(m,p),n)
            The leading p-by-n part of this array must contain the original
            state/output matrix C; the remainder of the leading max(m,p)-by-n
            part is used as internal workspace.
        D : input rank-2 array('d') with bounds (max(m,p),max(m,p))
            The leading p-by-m part of this array must contain the original
            direct transmission matrix D; the remainder of the leading
            max(m,p)-by-max(m,p) part is used as internal workspace.
        leri : input string(len=1)
            Indicates whether the left polynomial matrix representation or
            the right polynomial matrix representation is required.
    Optional arguments:
        equil := 'N' input string(len=1)
            Specifies whether the user wishes to balance the triplet (A,B,C),
            before computing a minimal state-space representation, as follows:
            = 'S':  Perform balancing (scaling);
            = 'N':  Do not perform balancing.
        tol := 0.0 input float
            The tolerance to be used in rank determination when transforming
            (A, B). If tol <= 0 a default value is used.
        ldwork := max(2*n+3*max(m,p), p*(p+2)) input int
            The length of the cache array.
            ldwork >= max( n + max(n, 3*m, 3*p), pm*(pm + 2))
            where pm = p, if leri = 'L';
                  pm = m, if leri = 'R'.
            For optimum performance it should be larger.
    Return objects:
        A_min : rank-2 array('d') with bounds (n,n)
            The leading nr-by-nr part of this array contains the upper block
            Hessenberg state dynamics matrix A_min of a minimal realization for
            the original system.
        B_min : rank-2 array('d') with bounds (n,max(m,p))
            The leading nr-by-m part of this array contains the transformed
            input/state matrix B_min.
        C_min : rank-2 array('d') with bounds (max(m,p),n)
            The leading p-by-nr part of this array contains the transformed
            state/output matrix C_min.
        nr : int
            The order of the minimal state-space representation
            (A_min,B_min,C_min).
        index : rank-1 array('i') with bounds either (p) or (m)
            If leri = 'L', index(i), i = 1,2,...,p, contains the maximum degree
            of the polynomials in the i-th row of the denominator matrix P(s)
            of the left polynomial matrix representation. These elements are
            ordered so that index(1) >= index(2) >= ... >= index(p).
            If leri = 'R', index(i), i = 1,2,...,m, contains the maximum degree
            of the polynomials in the i-th column of the denominator matrix P(s)
            of the right polynomial matrix representation. These elements are
            ordered so that index(1) >= index(2) >= ... >= index(m).
        pcoeff : rank-3 array('d') with bounds either (p,p,n+1) or (m,m,n+1)
            If leri = 'L' then porm = p, otherwise porm = m.
            The leading porm-by-porm-by-kpcoef part of this array contains
            the coefficients of the denominator matrix P(s), where
            kpcoef = max(index) + 1.
            pcoeff(i,j,k) is the coefficient in s**(index(iorj)-k+1) of
            polynomial (i,j) of P(s), where k = 1,2,...,kpcoef; if leri = 'L'
            then iorj = I, otherwise iorj = J. Thus for leri = 'L',
            P(s) = diag(s**index)*(pcoeff(.,.,1)+pcoeff(.,.,2)/s+...).
        qcoeff : rank-3 array('d') with bounds (p,m,n + 1) or (max(m,p),max(m,p))
            If leri = 'L' then porp = m, otherwise porp = p.
            If leri = 'L', the leading porm-by-porp-by-kpcoef part of this array
            contains the coefficients of the numerator matrix Q(s).
            If leri = 'R', the leading porp-by-porm-by-kpcoef part of this array
            contains the coefficients of the numerator matrix Q(s).
            qcoeff(i,j,k) is defined as for pcoeff(i,j,k).
        vcoeff : rank-3 array('d') with bounds (p,n,n+1) or (m,n,n+1)
            The leading porm-by-nr-by-kpcoef part of this array contains
            the coefficients of the intermediate matrix V(s).
            vcoeff(i,j,k) is defined as for pcoeff(i,j,k).
    Raises
    ------
    SlycotArithmeticError
        :info == 1:
            A singular matrix was encountered during the
            computation of V(s);
        :info == 2:
            A singular matrix was encountered during the
            computation of P(s).

    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['leri', 'equil', 'n', 'm', 'P', 'A', 'LDA'+hidden, 'B',
        'LDB'+hidden, 'C', 'LDC'+hidden, 'D', 'LDD'+hidden, 'nr', 'index',
        'pcoeff', 'LDPCO1'+hidden, 'LDPCO2'+hidden, 'qcoeff', 'LDQCO1'+hidden,
        'LDQCO2'+hidden, 'vcoeff', 'LDVCO1'+hidden, 'LDVCO2'+hidden, 'tol',
        'IWORK'+hidden, 'DWORK'+hidden, 'ldwork', 'INFO'+hidden]
    wfun = {"L": _wrapper.tb03ad_l,
            "R": _wrapper.tb03ad_r}
    mp_ = {"L": p, "R": m}
    mp = mp_[leri]
    if leri not in wfun.keys():
        raise SlycotParameterError('leri must be either L or R', -1)
    if ldwork is None:
        ldwork = max(2*n + 3*max(m, p), mp*(mp+2))
    out = wfun[leri](n, m, p, A, B, C, D, equil=equil, tol=tol, ldwork=ldwork)
    raise_if_slycot_error(out[-1], arg_list)
    return out[:-1]


def tb04ad(n,m,p,A,B,C,D,tol1=0.0,tol2=0.0,ldwork=None):
    """ Ar,Br,Cr,nr,denom_degs,denom_coeffs,num_coeffs = tb04ad(n,m,p,A,B,C,D,[tol1,tol2,ldwork])

    Convert a state-space system to a tranfer function or matrix of transfer functions.
    The transfer function is given as rows over common denominators.

    Required arguments
    ------------------

        n : integer
            state dimension
        m : integer
            input dimension
        p : integer
            output dimension
        A :  rank-2 array, shape(n,n)
            state dynamics matrix.
        B : rank-2 array, shape (n,m)
            input matrix
        C : rank-2 array, shape (p,n)
            output matri
        D : rank-2 array, shape (p,m)
            direct transmission matrix

     Optional arguments
     ------------------

        tol1 = 0.0: double
            tolerance in determining the transfer function coefficients,
            when set to 0, a default value is used
        tol2 = 0.0: double
            tolerance in separating out a controllable/observable subsystem
            of (A,B,C), when set to 0, a default value is used
        ldwork : int
            The length of the cache array. The default values is
            max(1,n*(n+1)+max(n*m+2*n+max(n,p),max(3*m,p)))

     Returns
     -------

        nr : int
            state dimension of the controllable subsystem
        Ar :  rank-2 array, shape(nr,nr)
            state dynamics matrix of the controllable subsystem
        Br : rank-2 array, shape (nr,m)
            input matrix of the controllable subsystem
        Cr : rank-2 array, shape (p,nr)
            output matri of the controllable subsystem
        index : rank-1 array, shape (p)
            array of orders of the denominator polynomials
        dcoeff : rank-2 array, shape (p,max(index)+1)
            array of denominator coefficients
        ucoeff : rank-3 array, shape (p,m,max(index)+1)
            array of numerator coefficients

    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['rowcol','n','m','p','A','lda'+hidden,'B','ldb'+hidden,'C','ldc'+hidden,'D', 'ldd'+hidden,
        'nr','index','dcoeff','lddcoe'+hidden, 'ucoeff','lduco1'+hidden,'lduco2'+hidden,'tol1','tol2','iwork'+hidden,'dwork'+hidden,'ldwork','info'+hidden]

    mp, pm = m, p
    porm, porp = p, m
    if ldwork is None:
        ldwork = max(1,n*(n+1)+max(n*mp+2*n+max(n,mp),3*mp,pm))
    if B.shape != (n, m):
        raise SlycotParameterError("The shape of B is ({}, {}), "
                                   "but expected ({}, {})"
                                   "".format(*(B.shape + (n, m))),
                                   -7)
    if C.shape != (p, n):
        raise SlycotParameterError("The shape of C is ({}, {}), "
                                   "but expected ({}, {})"
                                   "".format(*(C.shape + (p, n))),
                                   -9)
    if D.shape != (max(1, p), m):
        raise SlycotParameterError("The shape of D is ({}, {}), "
                                   "but expected ({}, {})"
                                   "".format(*(D.shape + (max(1, p), m))),
                                   -11)
    out = _wrapper.tb04ad_r(n,m,p,A,B,C,D,tol1,tol2,ldwork)

    raise_if_slycot_error(out[-1], arg_list)

    A,B,C,Nr,index,dcoeff,ucoeff = out[:-1]
    kdcoef = max(index)+1
    return A[:Nr,:Nr],B[:Nr,:m],C[:p,:Nr],Nr,index,dcoeff[:porm,:kdcoef],ucoeff[:porm,:porp,:kdcoef]


def tb05ad(n, m, p, jomega, A, B, C, job='NG'):
    """tb05ad(n, m, p, jomega, A, B, C, job='NG')

    To find the complex frequency response matrix (transfer matrix)
    G(freq) of the state-space representation (A,B,C) given by

    ::

                                  -1
       G(freq) = C * ((freq*I - A)  ) * B

    where A, B and C are real N-by-N, N-by-M and P-by-N matrices
    respectively and freq is a complex scalar.

    Parameters
    ----------
    n : int
        The number of states, i.e. the order of the state
        transition matrix A.
    m : int
        The number of inputs, i.e. the number of columns in the
        matrix B.
    p :  int
        The number of outputs, i.e. the number of rows in the
        matrix C.
    jomega : complex float
        The frequency at which the frequency response matrix
        (transfer matrix) is to be evaluated. For continuous time
        systems, this is j*omega, where omega is the frequency to
        be evaluated. For discrete time systems,
        freq = exp(j*omega*Ts)
    A : (n,n) ndarray
        On entry, this array must contain the state transition
        matrix A.
    B : (n,m) ndarray
        On entry, this array must contain the input/state matrix B.
    C : (p,n) ndarray
        On entry, of this array must contain the state/output matrix C.
    job : {'AG', 'NG', 'NH'}
        If job = 'AG' (i.e., 'all', 'general matrix'), the A matrix is
        first balanced. The balancing transformation
        is then appropriately applied to matrices B and C. The A matrix
        is (again) transformed to an upper Hessenberg representation and
        the B and C matrices are also transformed. In addition,
        the condition number of the problem is calculated as well as the
        eigenvalues of A.

        If job='NG' (i.e., 'none', 'general matrix'), no balancing is done.
        Neither the condition number nor the eigenvalues are calculated.
        The routine still transforms A into upper Hessenberg form. The
        matrices B and C are also appropriately transformed.

        If job = 'NH' (i.e., 'none', 'hessenberg matrix'), the function
        assumes the matrices have already been transformed into Hessenberg
        form, i.e., by a previous function call tb05ad. If this not the
        case, the routine will return a wrong result without warning.

    Returns
    -------
    if job = 'AG':
    --------------
      At:  The A matrix which has been both balanced and
           transformed to upper Hessenberg form. The balancing
           transforms A according to
                      A1 =   P^-1 * A * P.
           The transformation to upper Hessenberg form then yields
                      At = Q^T * (P^-1 * A * P ) * Q.
           Note that the lower triangle of At is in general not zero.
           Rather, it contains information on the orthogonal matrix Q
           used to transform A1 to Hessenberg form. See docs for lappack
           DGEHRD():
           http://www.netlib.org/lapack/explore-3.1.1-html/dgehrd.f.html
           However, it does not apparently contain information on P, the
           matrix used in the balancing procedure.

      Bt:  The matrix B transformed according to
                      Bt = Q^T * P^-1 * B.

      Ct:  The matrix C transformed according to
                      Ct = C * P * Q

      rcond: RCOND contains an estimate of the reciprocal of the
             condition number of matrix H with respect to inversion, where
                     H = (j*freq * I - A)

      g_jw: complex p-by-m array, which contains the frequency response
            matrix G(freq).

      ev:   Eigenvalues of the matrix A.

      hinvb : complex n-by-m array, which  contains the product
               -1
              H  B.

    if job = 'NG':
    --------------
      At:    The matrix A transformed to upper Hessenberg form according
             to
                      At = Q^T  * A  * Q.
             The lower triangle is not zero. It containts info on the
             orthoganal transformation. See docs for linpack DGEHRD()
             http://www.netlib.org/lapack/explore-3.1.1-html/dgehrd.f.html

      Bt:    The matrix B transformed according to
                      Bt = Q^T * B.

      Ct:    The matrix C transformed according to
                      Ct = C * Q
      g_jw:  complex array with dim p-by-m which contains the frequency
             response matrix G(freq).

      hinvb : complex array with dimension p-by-m.
              This array contains the
                      -1
             product H  B.

    if job = 'NH'
    --------------
      g_jw:  complex p-by-m array which contains the frequency
             response matrix G(freq).

      hinvb : complex p-by-m array which contains the
                      -1
             product H  B.


    Raises
    ------
    SlycotArithmeticError
        :info = 1:
            More than {n30} (30*`n`) iterations were required to isolate the
            eigenvalues of A. The computations are continued.
        :info = 2:
            Either `freq`={jomega} is too near to an eigenvalue of A,
            or `rcond` is less than the machine precision EPS.

    Example
    -------
    >>> A = np.array([[0.0, 1.0],
                [-100.0,   -20.1]])
    >>> B = np.array([[0.],[100]])
    >>> C = np.array([[1., 0.]])
    >>> n = np.shape(A)[0]
    >>> m = np.shape(B)[1]
    >>> p = np.shape(C)[0]
    >>> jw_s = [1j*10, 1j*15]
    >>> at, bt, ct, g_1, hinvb, info = slycot.tb05ad(n, m, p, jw_s[0],
                                                    A, B, C, job='NG')
    >>> g_2, hinv2,info = slycot.tb05ad(n, m, p, jw_s[1], at, bt, ct, job='NH')

    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['baleig'+hidden, 'inita'+hidden, 'n', 'm', 'p', 'freq', 'a',
                'lda'+hidden, 'b', 'ldb'+hidden, 'c', 'ldc'+hidden, 'rcond',
                'g', 'ldg'+hidden, 'evre', 'evim', 'hinvb', 'ldhinv'+hidden,
                'iwork'+hidden, 'dwork'+hidden, 'ldwork'+hidden,
                'zwork'+hidden, 'lzwork'+hidden, 'info'+hidden]
    # Fortran function prototype:
    # TB05AD(baleig,inita,n,m,p,freq,a,lda,b,ldb,c,ldc,rcond,g,ldg,evre,evim,hinvb,ldhinv,
    # iwork,dwork,ldwork,zwork,lzwork,info)

    # Sanity check on matrix dimensions
    if A.shape != (n, n):
        raise SlycotParameterError("The shape of A is ({0:}, {1:}), "
                                   "but expected ({2:}, {2:})"
                                   "".format(*(A.shape + (n,))),
                                   -7)
    if B.shape != (n, m):
        raise SlycotParameterError("The shape of B is ({0:}, {1:}), "
                                   "but expected ({2:}, {3:})"
                                   "".format(*(B.shape + (n, m))),
                                   -9)
    if C.shape != (p, n):
        raise SlycotParameterError("The shape of C is ({0:}, {1:}), "
                                   "but expected ({2:}, {3:})"
                                   "".format(*(C.shape + (p, n))),
                                   -11)

    # ----------------------------------------------------
    # Checks done, do computation.
    n30 = 30*n  # for INFO = 1 error docstring
    if job == 'AG':
        out = _wrapper.tb05ad_ag(n, m, p, jomega, A, B, C)
        At, Bt, Ct, rcond, g_jw, evre, evim, hinvb, info = out
        raise_if_slycot_error(info, arg_list, tb05ad.__doc__, locals())
        ev = _np.zeros(n, 'complex64')
        ev.real = evre
        ev.imag = evim
        return At, Bt, Ct, g_jw, rcond, ev, hinvb, info
    elif job == 'NG':
        # use tb05ad_ng, for 'NONE' , and 'General', because balancing
        # (option 'A' for 'ALL') seems to have  a bug.
        out = _wrapper.tb05ad_ng(n, m, p, jomega, A, B, C)
        At, Bt, Ct, g_jw, hinvb, info = out
        raise_if_slycot_error(info, arg_list, tb05ad.__doc__, locals())
        return At, Bt, Ct, g_jw, hinvb, info
    elif job == 'NH':
        out = _wrapper.tb05ad_nh(n, m, p, jomega, A, B, C)
        g_i, hinvb, info = out
        raise_if_slycot_error(info, arg_list, tb05ad.__doc__, locals())
        return g_i, hinvb, info
    else:
        raise SlycotParameterError("Unrecognized job. Expected job = 'AG' or "
                                   "job='NG' or job = 'NH' but received job={}"
                                   "".format(job),
                                   -1)  # job is baleig and inita together


def td04ad(rowcol,m,p,index,dcoeff,ucoeff,tol=0.0,ldwork=None):
    """ nr,A,B,C,D = td04ad(rowcol,m,p,index,dcoeff,ucoeff,[tol,ldwork])

    Convert a transfer function or matrix of transfer functions to
    a minimum state space realization.

    Parameters
    ----------
    rowcol : {R', 'C'}
        indicates whether the transfer matrix T(s) is given
        as rows ('R') or colums ('C') over common denominators.
    m : int
        input dimension
    p : int
        output dimension
    index : (p,) or (m,) array_like
        array of orders of the denominator polynomials. Different
        shapes corresponding to rowcol=='R' and rowcol=='C'
        respectively.
    dcoeff : (p,max(index)+1) or (m,max(index)+1) ndarray
        array of denominator coefficients. Different shapes
        corresponding to rowcol=='R' and rowcol=='C' respectively.
    ucoeff : (p,m,max(index)+1) or (max(p,m),max(p,m),max(index)+1) ndarray
        array of numerator coefficients. Different shapes
        corresponding to rowcol=='R' and rowcol=='C' respectively.
    tol : float, optional
        tolerance in determining the state space system,
        when set to 0, a default value is used.
    ldwork : int, optional
        The length of the cache array. The default values is
        max(1,sum(index)+max(sum(index),max(3*m,3*p)))

    Returns
    -------
    nr : int
        minimal state dimension
    A : (nr,nr) ndarray
        state dynamics matrix.
    B : (nr,m) ndarray
        input matrix
    C : (p,nr) ndarray
        output matrix
    D : (p,m) ndarray
        direct transmission matrix

    Raises
    ------
    SlycotArithmeticError
        :info > 0:
            i={info} is the first index of `dcoeff` for which
            ``abs( dcoeff(i,1) )`` is so small that the calculations
            would overflow (see SLICOT Library routine TD03AY);
            that is, the leading coefficient of a polynomial is
            nearly zero;
    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['rowcol','m','p','index','dcoeff','lddcoe'+hidden, 'ucoeff', 'lduco1'+hidden,'lduco2'+hidden,
        'nr','A','lda'+hidden,'B','ldb'+hidden,'C','ldc'+hidden,'D', 'ldd'+hidden,
        'tol','iwork'+hidden,'dwork'+hidden,'ldwork','info'+hidden]
    if ldwork is None:
        n = sum(index)
        ldwork = max(1,n+max(n,max(3*m,3*p)))

    kdcoef = max(index)+1
    if rowcol == 'R':
        if ucoeff.ndim != 3:
            raise SlycotParameterError("The numerator is not a 3D array!", -7)
        expectedshape = (max(1, p), max(1, m), kdcoef)
        if ucoeff.shape != expectedshape:
            raise SlycotParameterError("The numerator shape is ({}, {}, {}), "
                                       "but expected ({}, {}, {})".format(
                                           *(ucoeff.shape + expectedshape)),
                                       -7)
        expectedshape = (max(1, p), kdcoef)
        if dcoeff.shape != expectedshape:
            raise SlycotParameterError("The denominator shape is ({}, {}), "
                                       "but expected ({}, {})".format(
                                           *(dcoeff.shape + expectedshape)),
                                       -5)
        out = _wrapper.td04ad_r(m,p,index,dcoeff,ucoeff,n,tol,ldwork)
    elif rowcol == 'C':
        if ucoeff.ndim != 3:
            raise SlycotParameterError("The numerator is not a 3D array!", -7)
        expectedshape = (max(1, m, p), max(1, m, p), kdcoef)
        if ucoeff.shape != expectedshape:
            raise SlycotParameterError("The numerator shape is ({}, {}, {}), "
                                       "but expected ({}, {}, {})".format(
                                           *(ucoeff.shape + expectedshape)),
                                       -7)
        expectedshape = (max(1, m), kdcoef)
        if dcoeff.shape != expectedshape:
            raise SlycotParameterError("The denominator shape is ({}, {}), "
                                       "but expected ({}, {})".format(
                                           *(dcoeff.shape + expectedshape)),
                                       -5)
        out = _wrapper.td04ad_c(m,p,index,dcoeff,ucoeff,n,tol,ldwork)
    else:
        raise SlycotParameterError("Parameter rowcol had an illegal value", -1)

    raise_if_slycot_error(out[-1], arg_list, td04ad.__doc__)
    Nr, A, B, C, D = out[:-1]
    return Nr, A[:Nr,:Nr], B[:Nr,:m], C[:p,:Nr], D[:p,:m]

def tc04ad(m,p,index,pcoeff,qcoeff,leri,ldwork=None):
    """ n,rcond,a,b,c,d = tc04ad_l(m,p,index,pcoeff,qcoeff,leri,[ldwork])

    To find a state-space representation (A,B,C,D) with the same
    transfer matrix as that of a given left or right polynomial
    matrix representation, i.e.

        C*inv(sI-A)*B + D = inv(P(s))*Q(s)

    or

        C*inv(sI-A)*B + D = Q(s)*inv(P(s))

    respectively.

    Required arguments:
        m : input int
            The number of system inputs.  m > 0.
        p := len(index) input int
            The number of system outputs.  p > 0.
        index : input rank-1 array('i') with bounds (p) or (m)
            If leri = 'L', index(i), i = 1,2,...,p, must contain the maximum
            degree of the polynomials in the I-th row of the denominator matrix
            P(s) of the given left polynomial matrix representation.
            If leri = 'R', index(i), i = 1,2,...,m, must contain the maximum
            degree of the polynomials in the I-th column of the denominator
            matrix P(s) of the given right polynomial matrix representation.
        pcoeff : input rank-3 array('d') with bounds (p,p,*) or (m,m,*)
            If leri = 'L' then porm = p, otherwise porm = m. The leading
            porm-by-porm-by-kpcoef part of this array must contain
            the coefficients of the denominator matrix P(s). pcoeff(i,j,k) is
            the coefficient in s**(index(iorj)-K+1) of polynomial (I,J) of P(s),
            where k = 1,2,...,kpcoef and kpcoef = max(index) + 1; if leri = 'L'
            then iorj = i, otherwise iorj = j. Thus for leri = 'L',
            P(s) = diag(s**index)*(pcoeff(.,.,1)+pcoeff(.,.,2)/s+...).
            If leri = 'R', pcoeff is modified by the routine but restored on exit.
        qcoeff : input rank-3 array('d') with bounds (p,m,*) or (max(m,p),max(m,p),*)
            If leri = 'L' then porp = m, otherwise porp = p. The leading
            porm-by-porp-by-kpcoef part of this array must contain
            the coefficients of the numerator matrix Q(s).
            qcoeff(i,j,k) is defined as for pcoeff(i,j,k).
            If leri = 'R', qcoeff is modified by the routine but restored on exit.
        leri : input string(len=1)
            Indicates whether a left polynomial matrix representation or a right
            polynomial matrix representation is input as follows:
            = 'L':  A left matrix fraction is input;
            = 'R':  A right matrix fraction is input.
    Optional arguments:
        ldwork := max(m,p)*(max(m,p)+4) input int
            The length of the cache array. ldwork >= max(m,p)*(max(m,p)+4)
            For optimum performance it should be larger.
    Return objects:
        n : int
            The order of the resulting state-space representation.
            That is, n = sum(index).
        rcond : float
            The estimated reciprocal of the condition number of the leading row
            (if leri = 'L') or the leading column (if leri = 'R') coefficient
            matrix of P(s).
            If rcond is nearly zero, P(s) is nearly row or column non-proper.
        A : rank-2 array('d') with bounds (n,n)
            The leading n-by-n part of this array contains the state dynamics matrix A.
        B : rank-2 array('d') with bounds (n,max(m,p))
            The leading n-by-n part of this array contains the input/state matrix B;
            the remainder of the leading n-by-max(m,p) part is used as internal
            workspace.
        C : rank-2 array('d') with bounds (max(m,p),n)
            The leading p-by-n part of this array contains the state/output matrix C;
            the remainder of the leading max(m,p)-by-n part is used as internal
            workspace.
        D : rank-2 array('d') with bounds (max(m,p),max(m,p))
            The leading p-by-m part of this array contains the direct transmission
            matrix D; the remainder of the leading max(m,p)-by-max(m,p) part is
            used as internal workspace.
    Raises
    ------
    SlycotArithmeticError
        :info == 1 and leri = 'L':
            P(s) is not row proper
        :info == 1 and leri = 'R':
            P(s) is not column proper
    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['leri', 'm', 'P', 'index',
                'pcoeff', 'LDPCO1' + hidden, 'LDPCO2' + hidden,
                'qcoeff', 'LDQCO1' + hidden, 'LDQCO2' + hidden,
                'N', 'rcond',
                'A', 'LDA' + hidden, 'B', 'LDB' + hidden,
                'C', 'LDC' + hidden, 'D', 'LDD' + hidden,
                'IWORK' + hidden, 'DWORK' + hidden, 'ldwork',
                'INFO' + hidden]
    if ldwork is None:
        ldwork = max(m, p)*(max(m, p)+4)
    n = sum(index)
    wfun = {"L": _wrapper.tc04ad_l, "R": _wrapper.tc04ad_r}
    if leri not in wfun.keys():
        raise SlycotParameterError('leri must be either L or R', -1)
    out = wfun[leri](m, p, index, pcoeff, qcoeff, n)
    raise_if_slycot_error(out[-1], arg_list, tc04ad.__doc__, locals())
    return out[:-1]


def tc01od(m,p,indlin,pcoeff,qcoeff,leri):
    """ pcoeff,qcoeff = tc01od_l(m,p,indlim,pcoeff,qcoeff,leri)

    To find the dual right (left) polynomial matrix representation of a given
    left (right) polynomial matrix representation, where the right and left
    polynomial matrix representations are of the form Q(s)*inv(P(s)) and
    inv(P(s))*Q(s) respectively.

    Required arguments:
        m : input int
            The number of system inputs.  m > 0.
        p : input int
            The number of system outputs.  p > 0.
        indlim : input int
            The highest value of k for which pcoeff(.,.,k) and qcoeff(.,.,k)
            are to be transposed.
            k = kpcoef + 1, where kpcoef is the maximum degree of the polynomials
            in P(s).  indlim > 0.
        pcoeff : input rank-3 array('d') with bounds (p,p,indlim) or (m,m,indlim)
            If leri = 'L' then porm = p, otherwise porm = m.
            On entry, the leading porm-by-porm-by-indlim part of this array
            must contain the coefficients of the denominator matrix P(s).
            pcoeff(i,j,k) is the coefficient in s**(indlim-k) of polynomial
            (i,j) of P(s), where k = 1,2,...,indlim.
        qcoeff : input rank-3 array('d') with bounds (max(m,p),max(m,p),indlim)
            On entry, the leading p-by-m-by-indlim part of this array must
            contain the coefficients of the numerator matrix Q(s).
            qcoeff(i,j,k) is the coefficient in s**(indlim-k) of polynomial
            (i,j) of Q(s), where k = 1,2,...,indlim.
        leri : input string(len=1)
    Return objects:
        pcoeff : rank-3 array('d') with bounds (p,p,indlim)
            On exit, the leading porm-by-porm-by-indlim part of this array
            contains the coefficients of the denominator matrix P'(s) of
            the dual system.
        qcoeff : rank-3 array('d') with bounds (max(m,p),max(m,p),indlim)
            On exit, the leading m-by-p-by-indlim part of the array contains
            the coefficients of the numerator matrix Q'(s) of the dual system.
        info : int
            = 0:  successful exit;
            < 0:  if info = -i, the i-th argument had an illegal value.
    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['leri', 'M', 'P', 'indlim', 'pcoeff', 'LDPCO1'+hidden,
        'LDPCO2'+hidden, 'qcoeff', 'LDQCO1'+hidden, 'LDQCO2'+hidden,
        'INFO'+hidden]
    if leri == 'L':
        out = _wrapper.tc01od_l(m,p,indlin,pcoeff,qcoeff)
        raise_if_slycot_error(out[-1], arg_list)
        return out[:-1]
    if leri == 'R':
        out = _wrapper.tc01od_r(m,p,indlin,pcoeff,qcoeff)
        raise_if_slycot_error(out[-1], arg_list)
        return out[:-1]
    raise SlycotParameterError('leri must be either L or R', -1)

def tf01md(n,m,p,N,A,B,C,D,u,x0):
    """ xf,y = tf01md(n,m,p,N,A,B,C,D,u,x0)

    To compute the output sequence of a linear time-invariant
    open-loop system given by its discrete-time state-space model

    Required arguments:
        n : input int
            Order of the State-space representation.
        m : input int
            Number of inputs.
        p : input int
            Number of outputs.
        N : input int
            Number of output samples to be computed.
        A : input rank-2 array('d') with bounds (n,n)
            State dynamics matrix.
        B : input rank-2 array('d') with bounds (n,m)
            Input/state matrix.
        C : input rank-2 array('d') with bounds (p,n)
            State/output matrix.
        D : input rank-2 array('d') with bounds (p,m)
            Direct transmission matrix.
        u : input rank-2 array('d') with bounds (m,N)
            Input signal.
        x0 : input rank-1 array('d') with bounds (n)
            Initial state, at time 0.
    Return objects:
        xf : rank-1 array('d') with bounds (n)
            Final state, at time N+1.
        y : rank-2 array('d') with bounds (p,N)
            Output signal.
    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['n','m','p','ny','A','lda'+hidden,'B','ldb'+hidden,
        'C','ldc'+hidden,'D','ldd'+hidden,'u','ldu'+hidden,'x0',
        'y'+hidden,'ldy'+hidden,'dwork'+hidden,'info'+hidden]

    out = _wrapper.tf01md(n,m,p,N,A,B,C,D,u,x0)
    raise_if_slycot_error(out[-1], arg_list)
    return out[:-1]

def tf01rd(n,m,p,N,A,B,C,ldwork=None):
    """ H = tf01rd(n,m,p,N,A,B,C,[ldwork])

    To compute N Markov parameters M_1, M_2,..., M_N from the
    parameters (A,B,C) of a linear time-invariant system, where each
    M_k is an p-by-m matrix and k = 1,2,...,N.

    All matrices are treated as dense, and hence TF01RD is not
    intended for large sparse problems.


    Required arguments:
        n : input int
            Order of the State-space representation.
        m : input int
            Number of inputs.
        p : input int
            Number of outputs.
        N : input int
            Number of Markov parameters to be computed.
        A : input rank-2 array('d') with bounds (n,n)
            State dynamics matrix.
        B : input rank-2 array('d') with bounds (n,m)
            Input/state matrix.
        C : input rank-2 array('d') with bounds (p,n)
            State/output matrix.
    Optional arguments:
        ldwork := 2*na*nc input int
    Return objects:
        H : rank-2 array('d') with bounds (p,N*m)
            H[:,(k-1)*m : k*m] contains the k-th Markov parameter,
            for k = 1,2...N.
    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['n','m','p','N','A','lda'+hidden,'B','ldb'+hidden,'C',
        'ldc'+hidden,'H','ldh'+hidden,'dwork'+hidden,'ldwork','info'+hidden]

    if ldwork is None:
        out = _wrapper.tf01rd(n,m,p,N,A,B,C)
    else:
        out = _wrapper.tf01rd(n,m,p,N,A,B,C,ldwork=ldwork)
    raise_if_slycot_error(out[-1], arg_list)
    return out[0]

def tb01pd(n, m, p, A, B, C, job='M', equil='S', tol=1e-8, ldwork=None):
    """Ar, Br, Cr, nr = tb01pd(n,m,p,A,B,C,[job,equil,tol,ldwork])

    To find a reduced (controllable, observable, or minimal) state-
    space representation (Ar,Br,Cr) for any original state-space
    representation (A,B,C). The matrix Ar is in upper block
    Hessenberg form.

    Required arguments:
        n : input int
            Order of the State-space representation.
        m : input int
            Number of inputs.
        p : input int
            Number of outputs.
        A : input rank-2 array('d') with bounds (n,n)
            State dynamics matrix.
        B : input rank-2 array('d') with bounds (n,max(m,p))
            The leading n-by-m part of this array must contain the original
            input/state matrix B; the remainder of the leading n-by-max(m,p)
            part is used as internal workspace.
        C : input rank-2 array('d') with bounds (p,n)
            The leading p-by-n part of this array must contain the original
            state/output matrix C; the remainder of the leading max(1,m,p)-by-n
            part is used as internal workspace.
    Optional arguments:
        job : input char*1
            Indicates whether the user wishes to remove the
            uncontrollable and/or unobservable parts as follows:
            = 'M':  Remove both the uncontrollable and unobservable
                    parts to get a minimal state-space representation;
            = 'C':  Remove the uncontrollable part only to get a
                    controllable state-space representation;
            = 'O':  Remove the unobservable part only to get an
                    observable state-space representation.
        equil : input char*1
            Specifies whether the user wishes to preliminarily balance
            the triplet (A,B,C) as follows:
            = 'S':  Perform balancing (scaling);
            = 'N':  Do not perform balancing.
    Return objects:
        Ar : output rank-2 array('d') with bounds (nr,nr)
            Contains the upper block Hessenberg state dynamics matrix
            Ar of a minimal, controllable, or observable realization
            for the original system, depending on the value of JOB,
            JOB = 'M', JOB = 'C', or JOB = 'O', respectively.
        Br : output rank-2 array('d') with bounds (nr,m)
            Contains the transformed input/state matrix Br of a
            minimal, controllable, or observable realization for the
            original system, depending on the value of JOB, JOB = 'M',
            JOB = 'C', or JOB = 'O', respectively.  If JOB = 'C', only
            the first IWORK(1) rows of B are nonzero.
        Cr : output rank-2 array('d') with bounds (p,nr)

            Contains the transformed state/output matrix Cr of a
            minimal, C controllable, or observable realization for the
            original C system, depending on the value of JOB, JOB =
            'M', C JOB = 'C', or JOB = 'O', respectively.  C If JOB =
            'M', or JOB = 'O', only the last IWORK(1) columns C (in
            the first NR columns) of C are nonzero.
        nr : output int
            The order of the reduced state-space representation
            (Ar,Br,Cr) of a minimal, controllable, or observable
            realization for the original system, depending on
            JOB = 'M', JOB = 'C', or JOB = 'O'.
    """
    hidden = ' (hidden by the wrapper)'
    arg_list = ['job', 'equil', 'n','m','p','A','lda'+hidden,'B','ldb'+hidden,
                'C','ldc'+hidden,'nr','tol','iwork'+hidden,'dwork'+hidden,
                'ldwork','info'+hidden]
    if ldwork is None:
        ldwork = max(1, n+max(n,3*m,3*p))
    elif ldwork < max(1, n+max(n,3*m,3*p)):
        raise SlycotParameterError("ldwork is too small", -15)
    out = _wrapper.tb01pd(n=n,m=m,p=p,a=A,b=B,c=C,
                          job=job,equil=equil,tol=tol,ldwork=ldwork)

    raise_if_slycot_error(out[-1], arg_list)
    return out[:-1]

def tg01ad(l,n,m,p,A,E,B,C,thresh=0.0,job='A'):
    """ A,E,B,C,lscale,rscale = tg01ad(l,n,m,p,A,E,B,C,[thresh,job])

    To balance the matrices of the system pencil

            S =  ( A  B ) - lambda ( E  0 ) :=  Q - lambda Z,
                 ( C  0 )          ( 0  0 )

    corresponding to the descriptor triple (A-lambda E,B,C),
    by balancing. This involves diagonal similarity transformations
    (Dl*A*Dr - lambda Dl*E*Dr, Dl*B, C*Dr) applied to the system
    (A-lambda E,B,C) to make the rows and columns of system pencil
    matrices

                diag(Dl,I) * S * diag(Dr,I)

    as close in norm as possible. Balancing may reduce the 1-norms
    of the matrices of the system pencil S.

    The balancing can be performed optionally on the following
    particular system pencils

            S = A-lambda E,

            S = ( A-lambda E  B ),    or

            S = ( A-lambda E ).
                (     C      )
    Required arguments:
        l : input int
            The number of rows of matrices A, B, and E.  l >= 0.
        n : input int
            The number of columns of matrices A, E, and C.  n >= 0.
        m : input int
            The number of columns of matrix B.  m >= 0.
        p : input int
            The number of rows of matrix C.  P >= 0.
        A : rank-2 array('d') with bounds (l,n)
            The leading L-by-N part of this array must
            contain the state dynamics matrix A.
        E : rank-2 array('d') with bounds (l,n)
            The leading L-by-N part of this array must
            contain the descriptor matrix E.
        B : rank-2 array('d') with bounds (l,m)
            The leading L-by-M part of this array must
            contain the input/state matrix B.
            The array B is not referenced if M = 0.
        C : rank-2 array('d') with bounds (p,n)
            The leading P-by-N part of this array must
            contain the state/output matrix C.
            The array C is not referenced if P = 0.
    Optional arguments:
        job := 'A' input string(len=1)
          Indicates which matrices are involved in balancing, as
          follows:
          = 'A':  All matrices are involved in balancing;
          = 'B':  B, A and E matrices are involved in balancing;
          = 'C':  C, A and E matrices are involved in balancing;
          = 'N':  B and C matrices are not involved in balancing.
        thresh := 0.0 input float
            Threshold value for magnitude of elements:
            elements with magnitude less than or equal to
            THRESH are ignored for balancing. THRESH >= 0.
    Return objects:
        A : rank-2 array('d') with bounds (l,n)
          The leading L-by-N part of this array contains
          the balanced matrix Dl*A*Dr.
        E : rank-2 array('d') with bounds (l,n)
          The leading L-by-N part of this array contains
          the balanced matrix Dl*E*Dr.
        B : rank-2 array('d') with bounds (l,m)
          If M > 0, the leading L-by-M part of this array
          contains the balanced matrix Dl*B.
          The array B is not referenced if M = 0.
        C : rank-2 array('d') with bounds (p,n)
          If P > 0, the leading P-by-N part of this array
          contains the balanced matrix C*Dr.
          The array C is not referenced if P = 0.
        lscale : rank-1 array('d') with bounds (l)
          The scaling factors applied to S from left.  If Dl(j) is
          the scaling factor applied to row j, then
          SCALE(j) = Dl(j), for j = 1,...,L.
        rscale : rank-1 array('d') with bounds (n)
          The scaling factors applied to S from right.  If Dr(j) is
          the scaling factor applied to column j, then
          SCALE(j) = Dr(j), for j = 1,...,N.
    """

    hidden = ' (hidden by the wrapper)'
    arg_list = ['job', 'l', 'n', 'm', 'p', 'thresh', 'A', 'lda'+hidden, 'E','lde'+hidden,'B','ldb'+hidden,'C','ldc'+hidden, 'lscale', 'rscale', 'dwork'+hidden, 'info']

    A,E,B,C,lscale,rscale,info = _wrapper.tg01ad(job,l,n,m,p,thresh,A,E,B,C)
    raise_if_slycot_error(info, arg_list)
    return A,E,B,C,lscale,rscale

def tg01fd(l,n,m,p,A,E,B,C,Q=None,Z=None,compq='N',compz='N',joba='N',tol=0.0,ldwork=None):
    """ A,E,B,C,ranke,rnka22,Q,Z = tg01fd(l,n,m,p,A,E,B,C,[Q,Z,compq,compz,joba,tol,ldwork])

    To compute for the descriptor system (A-lambda E,B,C)
    the orthogonal transformation matrices Q and Z such that the
    transformed system (Q'*A*Z-lambda Q'*E*Z, Q'*B, C*Z) is
    in a SVD-like coordinate form with

                 ( A11  A12 )             ( Er  0 )
        Q'*A*Z = (          ) ,  Q'*E*Z = (       ) ,
                 ( A21  A22 )             (  0  0 )

    where Er is an upper triangular invertible matrix.
    Optionally, the A22 matrix can be further reduced to the form

                  ( Ar  X )
            A22 = (       ) ,
                  (  0  0 )

    with Ar an upper triangular invertible matrix, and X either a full
    or a zero matrix.
    The left and/or right orthogonal transformations performed
    to reduce E and A22 can be optionally accumulated.

    Required arguments:
        l : input int
            The number of rows of matrices A, B, and E.  l >= 0.
        n : input int
            The number of columns of matrices A, E, and C.  n >= 0.
        m : input int
            The number of columns of matrix B.  m >= 0.
        p : input int
            The number of rows of matrix C.  p >= 0.
        A : rank-2 array('d') with bounds (l,n)
            The leading l-by-n part of this array must
            contain the state dynamics matrix A.
        E : rank-2 array('d') with bounds (l,n)
            The leading l-by-n part of this array must
            contain the descriptor matrix E.
        B : rank-2 array('d') with bounds (l,m)
            The leading L-by-M part of this array must
            contain the input/state matrix B.
        C : rank-2 array('d') with bounds (p,n)
            The leading P-by-N part of this array must
            contain the state/output matrix C.
    Optional arguments:
        Q : rank-2 array('d') with bounds (l,l)
            If COMPQ = 'N':  Q is not referenced.
            If COMPQ = 'I':  Q need not be set.
            If COMPQ = 'U':  The leading l-by-l part of this
                            array must contain an orthogonal matrix
                            Q1.
        Z : rank-2 array('d') with bounds (n,n)
            If COMPZ = 'N':  Z is not referenced.
            If COMPZ = 'I':  Z need not be set.
            If COMPZ = 'U':  The leading n-by-n part of this
                            array must contain an orthogonal matrix
                            Z1.
        compq := 'N' input string(len=1)
            = 'N':  do not compute Q.
            = 'I':  Q is initialized to the unit matrix, and the
                    orthogonal matrix Q is returned.
            = 'U':  Q must contain an orthogonal matrix Q1 on entry,
                    and the product Q1*Q is returned.
        compz := 'N' input string(len=1)
            = 'N':  do not compute Z.
            = 'I':  Z is initialized to the unit matrix, and the
                    orthogonal matrix Z is returned.
            = 'U':  Z must contain an orthogonal matrix Z1 on entry,
                    and the product Z1*Z is returned.
        joba := 'N' input string(len=1)
            = 'N':  do not reduce A22.
            = 'R':  reduce A22 to a SVD-like upper triangular form.
            = 'T':  reduce A22 to an upper trapezoidal form.
        tol := 0 input float
            The tolerance to be used in determining the rank of E
            and of A22. If the user sets TOL > 0, then the given
            value of TOL is used as a lower bound for the
            reciprocal condition numbers of leading submatrices
            of R or R22 in the QR decompositions E * P = Q * R of E
            or A22 * P22 = Q22 * R22 of A22.
            A submatrix whose estimated condition number is less than
            1/TOL is considered to be of full rank.  If the user sets
            TOL <= 0, then an implicitly computed, default tolerance,
            defined by  TOLDEF = L*N*EPS,  is used instead, where
            EPS is the machine precision (see LAPACK Library routine
            DLAMCH). TOL < 1.
        ldwork : input int
            The length of the cache array.
            ldwork >= MAX( 1, n+p, MIN(l,n)+MAX(3*n-1,m,l) ).
            For optimal performance, ldwork should be larger.
    Return objects:
        A : rank-2 array('d') with bounds (l,n)
            On entry, the leading L-by-N part of this array must
            contain the state dynamics matrix A.
            On exit, the leading L-by-N part of this array contains
            the transformed matrix Q'*A*Z. If JOBA = 'T', this matrix
            is in the form

                         ( A11  *   *  )
                Q'*A*Z = (  *   Ar  X  ) ,
                         (  *   0   0  )

            where A11 is a RANKE-by-RANKE matrix and Ar is a
            RNKA22-by-RNKA22 invertible upper triangular matrix.
            If JOBA = 'R' then A has the above form with X = 0.
        E : rank-2 array('d') with bounds (l,n)
            The leading L-by-N part of this array contains
            the transformed matrix Q'*E*Z.

                     ( Er  0 )
            Q'*E*Z = (       ) ,
                     (  0  0 )

            where Er is a RANKE-by-RANKE upper triangular invertible
            matrix.
        B : rank-2 array('d') with bounds (l,m)
            The leading L-by-M part of this array contains
            the transformed matrix Q'*B.
        C : rank-2 array('d') with bounds (p,n)
            The leading P-by-N part of this array contains
            the transformed matrix C*Z.
        Q : rank-2 array('d') with bounds (l,l)
            If COMPQ = 'N':  Q is not referenced.
            If COMPQ = 'I':  The leading L-by-L part of this
                            array contains the orthogonal matrix Q,
                            where Q' is the product of Householder
                            transformations which are applied to A,
                            E, and B on the left.
            If COMPQ = 'U':  The leading L-by-L part of this
                            array contains the orthogonal matrix
                            Q1*Q.
        Z : rank-2 array('d') with bounds (n,n)
            If COMPZ = 'N':  Z is not referenced.
            If COMPZ = 'I':  The leading N-by-N part of this
                            array contains the orthogonal matrix Z,
                            which is the product of Householder
                            transformations applied to A, E, and C
                            on the right.
            If COMPZ = 'U':  The leading N-by-N part of this
                            array contains the orthogonal matrix
                            Z1*Z.
        ranke : output int
            The estimated rank of matrix E, and thus also the order
            of the invertible upper triangular submatrix Er.
        rnka22 : output int
            If JOBA = 'R' or 'T', then RNKA22 is the estimated rank of
            matrix A22, and thus also the order of the invertible
            upper triangular submatrix Ar.
            If JOBA = 'N', then RNKA22 is not referenced.
    """

    hidden = ' (hidden by the wrapper)'
    arg_list = ['compq', 'compz', 'joba', 'l', 'n', 'm', 'p', 'A', 'lda'+hidden, 'E','lde'+hidden,'B','ldb'+hidden,'C','ldc'+hidden,'Q','ldq'+hidden,'Z','ldz'+hidden,'ranke','rnka22','tol','iwork'+hidden, 'dwork'+hidden, 'ldwork', 'info']


    if compq != 'N' and compq != 'I' and compq != 'U':
        raise SlycotParameterError('Parameter compq had an illegal value', -1)

    if compz != 'N' and compz != 'I' and compz != 'U':
        raise SlycotParameterError('Parameter compz had an illegal value', -2)

    if joba != 'N' and joba != 'R' and joba != 'T':
        raise SlycotParameterError('Parameter joba had an illegal value', -3)

    if ldwork is None:
        ldwork = max(1, n+p, min(l,n) + max(3*n-1, m, l))


    if compq == 'N' and compz == 'N':
        A,E,B,C,ranke,rnka22,info = _wrapper.tg01fd_nn(joba,l,n,m,p,A,E,B,C,tol,ldwork)
        Q = None
        Z = None
    elif compq == 'I' and compz == 'I':
        A,E,B,C,Q,Z,ranke,rnka22,info = _wrapper.tg01fd_ii(joba,l,n,m,p,A,E,B,C,tol,ldwork)
    elif compq == 'U' and compz == 'U':
        A,E,B,C,Q,Z,ranke,rnka22,info = _wrapper.tg01fd_uu(joba,l,n,m,p,A,E,B,C,Q,Z,tol,ldwork)
    else:
        raise NotImplementedError(
            "The combination of compq and compz is not implemented")

    raise_if_slycot_error(info, arg_list)

    if joba == 'N':
        rnka22 = None

    return A,E,B,C,ranke,rnka22,Q,Z

# to be replaced by python wrappers



#
# test_transform.py - generic tests for transform routines
# repagh <rene.vanpaassen@gmail.com, May 2020

import pytest
from .test_exceptions import assert_docstring_parse
from slycot import transform as tf
from slycot.exceptions import SlycotArithmeticError


@pytest.mark.parametrize(
    'fun,        exception_class,       erange,         checkvars',
    ((tf.tb03ad, SlycotArithmeticError, 2,              {}),
     (tf.tb05ad, SlycotArithmeticError, 2,              {'n30': 90,
                                                         'jomega': 2.0,
                                                         'rcond': 1e-12}),
     (tf.td04ad, SlycotArithmeticError, (3,),           {}),
     (tf.tc04ad, SlycotArithmeticError, 1,              {'leri': 'L'})))
def test_transform_docparse(fun, exception_class, erange, checkvars):
    assert_docstring_parse(fun.__doc__,  exception_class, erange, checkvars)



"""

test_examples.py


"""

from inspect import getmembers, isfunction
import pytest

from slycot import examples

examplefunctions = [fun for (fname, fun) in getmembers(examples)
                    if isfunction(fun) and "_example" in fname]

# Exempt certain functions to produce warnings with attributes (iwarn, info)
#
# Failed to compute beta(A) within the specified tolerance
examples.ab13fd_example.allow_iwarninfo = [(None, 1)]


def check_warn(recwarn, examplefun):
    """Returns True if a warning occurs that is not exempt"""
    for w in recwarn:
        try:
            if (w.message.iwarn, w.message.info) in examplefun.allow_iwarninfo:
                continue
        except AttributeError:
            pass
        return True


@pytest.mark.parametrize('examplefun', examplefunctions)
#ignore numpy ABI change warnings https://github.com/numpy/numpy/pull/432
@pytest.mark.filterwarnings("ignore:numpy.dtype size changed")
@pytest.mark.filterwarnings("ignore:numpy.ufunc size changed")
def test_example(examplefun, capsys, recwarn):
    """
    Test the examples.

    Test that all the examples work, produce some (unchecked) output but no
    exceptions or warnings.
    """
    examplefun()
    captured = capsys.readouterr()

    # fail for first in order
    failconditions = [
        ((not len(captured.out) > 0), "Example {} did not print any results\n"),
        (captured.err, "Example {} wrote to stderr\n"),
        (check_warn(recwarn, examplefun), "Example {} produced a warning.\n")]
    for failed, msgfmt  in failconditions:
        if failed:
            pytest.fail(msgfmt.format(examplefun.__name__) +
                        "Captured output:\n{}\n"
                        "Captured stderr:\n{}\n"
                        "Captured warnings:\n{}\n"
                        "".format(captured.out,
                                  captured.err,
                                  [w.message for w in recwarn]))




# THIS FILE IS GENERATED FROM SLYCOT SETUP.PY
short_version = '0.4.0.0'
version = '0.4.0.0'
full_version = '0.4.0.0.git'
git_revision = ''
release = True

if not release:
        version = full_version



try:
    __SLYCOT_SETUP__
except NameError:
    __SLYCOT_SETUP__ = False


if __SLYCOT_SETUP__:
    import sys as _sys
    _sys.stderr.write('Running from numpy source directory.\n')
    del _sys
else:

    # import slycot.examples

    # Analysis routines (15/40 wrapped)
    from .analysis import ab01nd, ab05md, ab05nd, ab07nd, ab08nd, ab08nz
    from .analysis import ab09ad, ab09ax, ab09bd, ab09md, ab09nd
    from .analysis import ab13bd, ab13dd, ab13ed, ab13fd


    # Data analysis routines (0/7 wrapped)

    # Filtering routines (0/6 wrapped)

    # Identification routines (0/5 wrapped)

    # Mathematical routines (7/81 wrapped)
    from .math import mc01td, mb03rd, mb03vd, mb03vy, mb03wd, mb05md, mb05nd

    # Synthesis routines (15/50 wrapped)

    from .synthesis import sb01bd,sb02md,sb02mt,sb02od,sb03md,sb03od
    from .synthesis import sb04md,sb04qd,sb10ad,sb10dd,sb10hd,sg03ad
    from .synthesis import sg02ad, sg03bd, sb10fd

    # Transformation routines (9/40 wrapped)
    from .transform import tb01id, tb03ad, tb04ad
    from .transform import tb05ad
    from .transform import tc04ad, tc01od
    from .transform import tf01md, tf01rd
    from .transform import td04ad, tb01pd

    # Version information
    from .version import version as __version__

    from numpy.testing import Tester
    test = Tester().test
    bench = Tester().bench



# ===================================================
# ab08n* tests

import unittest
from slycot import analysis
import numpy as np

from scipy.linalg import eig
from numpy.testing import assert_equal, assert_allclose


class test_ab08nX(unittest.TestCase):
    """ Test regular pencil construction ab08nX with input parameters
    according to example in documentation """

    A = np.diag([1., 1., 3., -4., -1., 3.])

    B = np.array([[ 0., -1.],
                  [-1.,  0.],
                  [ 1., -1.],
                  [ 0.,  0.],
                  [ 0.,  1.],
                  [-1., -1.]])

    C = np.array([[1., 0., 0., 1., 0., 0.],
                  [0., 1., 0., 1., 0., 1.],
                  [0., 0., 1., 0., 0., 1.]])

    D = np.zeros((3, 2))

    def normalize(self, w):
        wi = np.flip(np.argsort(np.abs(w)))
        wn = w[wi]/w[wi[0]]
        return wn

    def ab08nX(self, ab08fun, A, B, C, D):
        n = 6
        m = 2
        p = 3
        # Check the observability and compute the ordered set of
        # the observability indices (call the routine with M = 0).
        out = ab08fun(n, 0, p, A, B, C, D)
        nu, rank, dinfz, nkror, nkrol, infz, kronr, kronl, Af, Bf = out[:10]

        assert_equal(kronl[:nkrol], np.array([1, 2, 2]))
        assert_equal(n-nu, 5)
        assert_allclose(Af[:nu, :nu], np.array([[-1.]]))
        # Check the controllability and compute the ordered set of
        # the controllability indices (call the routine with P = 0)
        out = ab08fun(n, m, 0, A, B, C, D)
        nu, rank, dinfz, nkror, nkrol, infz, kronr, kronl, Af, Bf = out[:10]
        assert_equal(kronr[:nkror], np.array([2, 3]))
        assert_equal(n-nu, 5)
        assert_allclose(Af[:nu, :nu], np.array([[-4.]]))
        # Compute the structural invariants of the given system.
        out = ab08fun(n, m, p, A, B, C, D)
        nu, rank, dinfz, nkror, nkrol, infz, kronr, kronl, Af, Bf = out[:10]
        assert_equal(nu, 2)
        # Compute the invariant zeros of the given system.
        w = eig(Af[:nu, :nu], Bf[:nu, :nu], left=False, right=False)
        w_ref = np.array([-2., 1.])
        assert_allclose(self.normalize(w), self.normalize(w_ref))
        # the examples value of infinite zeros does not match the code
        # compare output formats to given strings
        # assert_equal(sum(infz[:dinfz]), 2)
        # assert_equal([[infz[i], i+1] for i in range(dinfz)], [[1, 1]])
        assert_equal(nkror, 0)
        assert_equal(nkrol, 1)
        assert_equal(kronl[:nkrol], np.array([2]))

    def test_ab08nd(self):
        "Test Construct regular pencil for real matrices"
        self.ab08nX(analysis.ab08nd, self.A, self.B, self.C, self.D)

    def test_ab08nz(self):
        "Test Construct regular pencil for (pseudo) complex matrices"
        Ac, Bc, Cc, Dc = [M.astype(np.complex128) for M in [self.A, self.B,
                                                            self.C, self.D]]
        self.ab08nX(analysis.ab08nz, Ac, Bc, Cc, Dc)


if __name__ == "__main__":
    unittest.main()



"""
Test ab01 wrappers

@author: bnavigator
"""

from numpy import array
from numpy.testing import assert_allclose, assert_equal

from scipy.linalg.lapack import dorgqr

from slycot.analysis import ab01nd


def test_ab01nd():
    """SLICOT doc example

    http://slicot.org/objects/software/shared/doc/AB01ND.html"""

    # Example program data
    n = 3
    m = 2
    tol = 0.0

    A = array([[-1., 0., 0.],
               [-2., -2., -2.],
               [-1., 0., -3.]])
    B = array([[1., 0.],
               [0,  2.],
               [0., 1.]])

    for jobz in ['N', 'I', 'F']:
        Ac, Bc, ncont, indcon, nblk, Z, tau = ab01nd(n, m, A, B,
                                                     jobz=jobz, tol=tol)

        # The transformed state dynamics matrix of a controllable realization
        assert_allclose(Ac[:ncont, :ncont], array([[-3.0000,  2.2361],
                                                   [ 0.0000, -1.0000]]),
                        atol=0.0001)

        # and the dimensions of its diagonal blocks are
        assert_equal(nblk[:indcon], array([2]))

        # The transformed input/state matrix B of a controllable realization
        assert_allclose(Bc[:ncont, :],array([[ 0.0000, -2.2361],
                                             [ 1.0000,  0.0000]]),
                        atol=0.0001)

        # The controllability index of the transformed system representation
        assert indcon == 1

        if jobz == 'N':
            assert Z is None
            continue
        elif jobz == 'I':
            Z_ = Z
        elif jobz == 'F':
            Z_, _, info = dorgqr(Z, tau)
            assert info == 0

        # The similarity transformation matrix Z
        assert_allclose(Z_, array([[ 0.0000,  1.0000,  0.0000],
                                  [-0.8944,  0.0000, -0.4472],
                                  [-0.4472,  0.0000,  0.8944]]),
                        atol=0.0001)




#
# test_mc.py - test suite for polynomial and rational function manipulation
# bnavigator <code@bnavigator.de>, Aug 2019

import pytest
import re

from slycot import mc01td
from slycot.exceptions import SlycotResultWarning


def test_mc01td():
    """ test_mc01td: doc example
    data from http://slicot.org/objects/software/shared/doc/MC01TD.html
    """
    (dp, stable, nz) = mc01td('C', 4, [2, 0, 1, -1, 1])
    assert dp == 4
    assert stable == 0
    assert nz == 2

def test_mc01td_D():
    """ test_mc01td_D: test discrete option """
    (dp, stable, nz) = mc01td('D', 3, [1, 2, 3, 4])
    assert dp == 3
    assert stable == 1
    assert nz == 0
    (dp, stable, nz) = mc01td('D', 3, [4, 3, 2, 1])
    assert dp == 3
    assert stable == 0
    assert nz == 3

def test_mc01td_warnings():
    """ test_mc01td_warnings: Test warnings """
    T = [([0, 0], "Entry ``P(x)`` is the zero polynomial."),
         ([0, 1], "The polynomial ``P(x)`` is most probably unstable,\n"
                  "although it may be stable with one or more zeros\n"
                  "very close to the imaginary axis.\n"
                  "The number of unstable zeros (NZ) is not determined."),
         ([1, 0], "The degree of the polynomial ``P(x)`` has been\n"
                  "reduced to ``(DB - 1)`` because\n"
                  "``P(DB+1-j) = 0.0`` on entry\n"
                  "for ``j = 0, 1,..., k-1`` and ``P(DB+1-k) <> 0.0``.")]
    for P, m in T:
        with pytest.warns(SlycotResultWarning, match=re.escape(m)):
            (dp, stable, nz) = mc01td('C', len(P)-1, P)



#
# test_sg03ad.py - test suite for stability margin commands
# RvP, 15 Jun 2017
from __future__ import print_function

import unittest
from slycot import synthesis
import numpy as np

from numpy.testing import assert_almost_equal

# test cases from
# Penzl T., Numerical Solution of Generalized Lyapunov Equations
# http://www.qucosa.de/fileadmin/data/qucosa/documents/4168/data/b002.pdf


class test_sg03ad(unittest.TestCase):

    def test_sg03ad_ex1c(self):
        """ Example 1 continuous case"""
        n = 100
        Xref = np.ones((n, n))
        U = np.tril(Xref)
        for t in range(0, 50, 10):
            A = (2**(-t) - 1) * np.eye(n) + np.diag(np.arange(1., n+1.)) + U.T
            E = np.eye(n) + 2**(-t) * U
            Y = A.T.dot(Xref).dot(E) + E.T.dot(Xref).dot(A)
            Q = np.zeros((n, n))
            Z = np.zeros((n, n))
            A, E, Q, Z, X, scale, sep, ferr, alphar, alphai, beta = \
                synthesis.sg03ad('C', 'B', 'N', 'N', 'L', n, A, E, Q, Z, Y)
            assert_almost_equal(X, Xref)

    def test_sg03ad_ex1d(self):
        """ Example 1 discrete case"""
        n = 100
        Xref = np.ones((n, n))
        U = np.tril(Xref)
        for t in range(0, 50, 10):
            A = 2**(-t) * np.eye(n) + np.diag(np.arange(1., n+1.)) + U.T
            E = np.eye(n) + 2**(-t) * U
            Y = A.T.dot(Xref).dot(A) - E.T.dot(Xref).dot(E)
            Q = np.zeros((n, n))
            Z = np.zeros((n, n))
            A, E, Q, Z, X, scale, sep, ferr, alphar, alphai, beta = \
                synthesis.sg03ad('D', 'B', 'N', 'N', 'L', n, A, E, Q, Z, Y)
            assert_almost_equal(X, Xref)

    def test_sg03ad_b1(self):
        """ SLICOT doc example / Penzl B.1 """
        n = 3
        A = np.array([[3.0, 1.0, 1.0],
                      [1.0, 3.0, 0.0],
                      [1.0, 0.0, 2.0]])
        E = np.array([[1.0, 3.0, 0.0],
                      [3.0, 2.0, 1.0],
                      [1.0, 0.0, 1.0]])
        Y = np.array([[64.0, 73.0, 28.0],
                      [73.0, 70.0, 25.0],
                      [28.0, 25.0, 18.0]])
        Xref = np.array([[-2.0000, -1.0000, 0.0000],
                         [-1.0000, -3.0000, -1.0000],
                         [0.0000, -1.0000, -3.0000]])
        Q = np.zeros((3, 3))
        Z = np.zeros((3, 3))
        A, E, Q, Z, X, scale, sep, ferr, alphar, alphai, beta = \
            synthesis.sg03ad('C', 'B', 'N', 'N', 'L', n, A, E, Q, Z, -Y)
        # print(A, E, Q, Z, X, scale, sep)
        assert_almost_equal(X, Xref)


if __name__ == "__main__":
    unittest.main()



#
# test_ab.py - generic tests for analysis programs
# repagh <rene.vanpaassen@gmail.com, May 2020

import pytest
from .test_exceptions import assert_docstring_parse
from slycot import analysis
from slycot.exceptions import SlycotArithmeticError, SlycotResultWarning


@pytest.mark.parametrize(
    'fun,              exception_class,       erange,         checkvars',
    ((analysis.ab05nd, SlycotArithmeticError, 1,              {'p1': 1}),
     (analysis.ab07nd, SlycotResultWarning,   2,              {'m': 1}),
     (analysis.ab09ad, SlycotArithmeticError, 3,              {'dico': 'C'}),
     (analysis.ab09ad, SlycotArithmeticError, (2,),           {'dico': 'D'}),
     (analysis.ab09ad, SlycotResultWarning,   ((1, 0), ),     {'nr': 3,
                                                               'Nr': 2}),
     (analysis.ab09ax, SlycotArithmeticError, 2,              {'dico': 'C'}),
     (analysis.ab09ax, SlycotResultWarning,   ((1, 0), ),     {'nr': 3,
                                                               'Nr': 2}),
     (analysis.ab09ad, SlycotArithmeticError, 3,              {'dico': 'C'}),
     (analysis.ab09ad, SlycotResultWarning,   ((1, 0), ),     {'nr': 3,
                                                               'Nr': 2}),
     (analysis.ab09md, SlycotArithmeticError, 3,              {'alpha': -0.1}),
     (analysis.ab09md, SlycotResultWarning,   ((1, 0), (2, 0)), {'nr': 3,
                                                               'Nr': 2,
                                                               'alpha': -0.1}),
     (analysis.ab09nd, SlycotArithmeticError, 3,              {'alpha': -0.1}),
     (analysis.ab09nd, SlycotResultWarning,   ((1, 0), (2, 0)), {'nr': 3,
                                                               'Nr': 2,
                                                               'alpha': -0.1}),
     (analysis.ab13bd, SlycotArithmeticError, 6,              {'dico': 'C'}),
     (analysis.ab13bd, SlycotResultWarning,   ((1, 0),),      {}),
     (analysis.ab13dd, SlycotArithmeticError, 4,              {}),
     (analysis.ab13ed, SlycotArithmeticError, 1,              {}),
     (analysis.ab13fd, SlycotArithmeticError, (2,),           {}),
     (analysis.ab13fd, SlycotResultWarning,   (1,),           {})))
def test_ab_docparse(fun, exception_class, erange, checkvars):
    assert_docstring_parse(fun.__doc__,  exception_class, erange, checkvars)





"""
exceptions.py

Copyright 2020 Slycot team

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
MA 02110-1301, USA.
"""

import re

from warnings import warn


class SlycotError(RuntimeError):
    """Slycot exception base class"""

    def __init__(self, message, info):
        super(SlycotError, self).__init__(message)
        self.info = info


class SlycotParameterError(SlycotError, ValueError):
    """A Slycot input parameter had an illegal value.

    In case of a wrong input value, the SLICOT routines return a negative
    info parameter indicating which parameter was illegal.
    """

    pass


class SlycotArithmeticError(SlycotError, ArithmeticError):
    """A Slycot computation failed"""

    pass


class SlycotWarning(UserWarning):
    """Slycot Warning"""

    def __init__(self, message, iwarn, info):
        super(SlycotWarning, self).__init__(message)
        self.info = info
        self.iwarn = iwarn


class SlycotResultWarning(SlycotWarning):
    """Slycot computation result warning

    A Slycot routine returned a nonzero info parameter that warns about the
    returned results, but the results might still be usable.
    """

    pass


def _parse_docsection(section_name, docstring, checkvars):
    slycot_error = None
    message = None
    docline = iter(docstring.splitlines())
    try:

        info_eval = False
        while section_name not in next(docline):
            continue
        section_indent = next(docline).index("-")

        for l in docline:
            # ignore blank lines
            if not l.strip():
                continue

            # reached next section without match
            if l[section_indent] == "-":
                break

            # Exception Type
            ematch = re.match(
                r'(\s*)(Slycot.*(Error|Warning))', l)
            if ematch:
                error_indent = len(ematch.group(1))
                slycot_error = ematch.group(2)

            # new infospec
            if slycot_error:
                imatch = re.match(
                    r'(\s{' + str(error_indent + 1) + r',}):(.+):\s*(.*)', l)
                if imatch:
                    infospec_indent = len(imatch.group(1))
                    infospec = imatch.group(2)
                    # Don't handle the standard case unless we have i
                    if infospec == "info = -i":
                        if 'i' not in checkvars.keys():
                            continue
                    infospec_ = infospec.replace(" = ", " == ")
                    try:
                        info_eval = eval(infospec_, checkvars)
                    except NameError:
                        raise RuntimeError("Unknown variable in infospec: "
                                           + infospec)
                    except SyntaxError:
                        raise RuntimeError("Invalid infospec: " + infospec)
                    if info_eval:
                        message = imatch.group(3).strip() + '\n'
                        mmatch = re.match(
                            r'(\s{' + str(infospec_indent+1) + r',})(.*)',
                            next(docline))
                        if not mmatch:
                            break  # docstring
                        body_indent = len(mmatch.group(1))
                        message += mmatch.group(2) + '\n'
                        for l in docline:
                            if l and not l[:body_indent].isspace():
                                break  # message body
                            message += l[body_indent:] + '\n'
                        break  # docstring
    except StopIteration:
        pass
    return (slycot_error, message)


def raise_if_slycot_error(info, arg_list=None, docstring=None, checkvars=None):
    """Raise exceptions or warnings if slycot info returned is non-zero.

    Parameters
    ----------
    info: int or list of int
        The parameter INFO or [IWARN, INFO] returned by the SLICOT subroutine
    arg_list: list of str, optional
        A list of arguments (possibly hidden by the wrapper) of the SLICOT
        subroutine
    docstring: str, optional
        The docstring of the Slycot function
    checkvars: dict, optional
        dict of variables for evaluation of <infospec> and formatting the
        exception message

    Notes
    -----
    If the numpydoc compliant docstring has a "Raises" section with one or
    multiple definition terms ``SlycotError`` or a subclass of it,
    the matching exception text is used.

    To raise warnings, define a "Warns" section using a ``SlycotWarning``
    definition or a subclass of it.

    The definition body must contain a reST compliant field list with
    ':<infospec>:' as field name, where <infospec> is a python parseable
    expression using the arguments `iwarn`, `info` and any additional variables
    provided in `checkvars` (usually obtained by calling `locals()`.
    A single " = " is treated as " == ".

    The body of the field list contains the exception or warning message and
    can contain replacement fields in format string syntax using the variables
    in `checkvars`.

    For negative info, the argument as indicated in arg_list was erroneous and
    a generic SlycotParameterError is raised if matching infospec was defined.

    Example
    -------
    >>> def fun(info):
    ...     '''Example function
    ...
    ...     Raises
    ...     ------
    ...     SlycotArithmeticError
    ...         :info = 1: INFO is 1
    ...         :info > 1 and info < n:
    ...             INFO is {info}, which is between 1 and {n}
    ...         :n <= info < m:
    ...             {info} is in [{n}, {m:10.2g})!
    ...
    ...     Warns
    ...     -----
    ...     SlycotResultWarning
    ...         :info >= 120: {info} is too large
    ...     SlycotResultWarning
    ...         :iwarn == 1: IWARN is 1
    ...     '''
    ...     n, m = 4, 120.
    ...     raise_if_slycot_error(info,
    ...                           arg_list=["a", "b", "c"],
    ...                           docstring=(fun.__doc__ if type(info) is list
    ...                                          else fun.__doc__[:-60]),
    ...                           checkvars=locals())
    ...
    >>> fun(0)
    >>> fun(-1)
    SlycotParameterError:
    The following argument had an illegal value: a
    >>> fun(1)
    SlycotArithmeticError:
    INFO is 1
    >>> fun(2)
    SlycotArithmeticError:
    INFO is 2, which is between 1 and 4
    >>> fun(4)
    SlycotArithmeticError:
    4 is in [4,    1.2e+02)!
    >>> fun(120)
    SlycotResultWarning:
    120 is too large
    >>> fun([1,0])
    SlycotResultWarning:
    IWARN is 1
    """
    try:
        iwarn, info = info
    except TypeError:
        iwarn = None
    if not checkvars:
        checkvars = {}
    if docstring and (iwarn or info):
        # possibly override info with mandatory argument
        checkvars['info'] = info
        # do not possibly override iwarn if not provided
        if iwarn is not None:
            checkvars['iwarn'] = iwarn

        exception, message = _parse_docsection("Raises", docstring, checkvars)
        if exception and message:
            fmessage = '\n' + message.format(**checkvars).strip()
            raise globals()[exception](fmessage, info)

        warning, message = _parse_docsection("Warns", docstring, checkvars)
        if warning and message:
            fmessage = '\n' + message.format(**checkvars).strip()
            warn(globals()[warning](fmessage, iwarn, info))
            return

    if info < 0 and arg_list:
        message = ("The following argument had an illegal value: {}"
                   "".format(arg_list[-info-1]))
        raise SlycotParameterError(message, info)

    # catch all
    if info > 0:
        raise SlycotError("Caught unhandled nonzero INFO value {}"
                          "".format(info),
                          info)
    if iwarn is None and 'iwarn' in checkvars:
        iwarn = checkvars['iwarn']
    if iwarn:
        warn(SlycotWarning("Caught unhandled nonzero IWARN value {}"
                           "".format(iwarn),
                           iwarn, info))
